/*! \file
    \brief ICP code for JM60

    \verbatim
    Opensource BDM  - interface DLL
    Based on TBDML interface DLL by
    Copyright (C) 2005  Daniel Malik

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    \endverbatim

*/
#include <windows.h>
#include <stdio.h>
#ifdef TBDML
#define _opensourcebdm_open         _tbdml_open
#define _opensourcebdm_close        _tbdml_close
#define _opensourcebdm_get_version  _tbdml_get_version
#define _opensourcebdm_init         _tbdml_init
#endif
#include "osbdm.h"
#include "common_usb.h"
#include "Configure.h"
#include "log.h"
#include "version.h"
#include "Common.h"
#include "ICP.h"
#include "Resource.h"

#include "FlashImage.h" // Contains the BDM Firmware

static const char *BDMCPUTypes[]={"JB16","????","JM16","UF32"};
static const char *hardwareDescription[] = {
   "Unknown or invalid hardware",                  // 0
   "USBDM - Extended TBDML/OSBDM (JB16DWE)",       // 1
   "TBDML - Minimal TBDML",                        // 2
   "TBDML - Minimal TBDML (Internal version)",     // 3
   "OSBDM - Original OSBDM",                       // 4
   "Witztronics TBDML/OSBDM",                      // 5
   "OSBDM+ - Extended OSBDM (RS08 support)",       // 6
   "USBDM - Extended TBDML/OSBDM (JMxxCLD)",       // 7
   "USBDM - Extended TBDML/OSBDM (JMxxCLC)",       // 8
   "USBSPYDER08 - SofTec USBSPYDER08      ",       // 9
   "USBDM - Extended TBDML/OSBDM (UF32PBE)",       // 10
   "Reserved #1",                                  // 11
   "Reserved #2",                                  // 12
   "Reserved #3",                                  // 13
   "Reserved #4",                                  // 14
   "Custom",                                       // 15
   };



// Version information read from BDM
U8   currentSoftwareVersion;
U8   currentHardwareVersion;
// Version information available in driver DLL
U8   availableSoftwareVersion;
U8   availableHardwareVersion;

//! \brief Windows Callback to handle dialogue events
//!
//! @param hDlg     Window Handle for the dialogue
//! @param message  The usual windows stuff
//! @param wParam
//! @param lParam
//!
static LRESULT CALLBACK Configure(HWND hDlg, UINT message,
                                  WPARAM wParam, LPARAM lParam)
{
   switch (message) {
      case WM_INITDIALOG: {
         char buff[100];
         U8 hardwareVersion = currentHardwareVersion;

         // Description of hardware
         if ((hardwareVersion & 0x3F) >
             sizeof(hardwareDescription)/sizeof(hardwareDescription[0])) {
            hardwareVersion = 0;
         }

         SetWindowText( GetDlgItem(hDlg, IDC_HARDWARE_TYPE),
                        hardwareDescription[hardwareVersion&0x3F]);

         // Description of current software in BDM
         sprintf(buff,"%s-%d.%d",
                 BDMCPUTypes[(currentHardwareVersion>>6)&0x03],
                 (currentSoftwareVersion>>4)&0xF,currentSoftwareVersion&0xF);
         SetWindowText( GetDlgItem(hDlg, IDC_CURRENT_VER), buff);

         // Description of version available in driver
         sprintf(buff,"%s-%d.%d",
                 BDMCPUTypes[(currentHardwareVersion>>6)&0x03],
                 (availableSoftwareVersion>>4)&0xF,availableSoftwareVersion&0xF);
         SetWindowText( GetDlgItem(hDlg, IDC_NEW_VER), buff);
         return TRUE;
      }

      case WM_COMMAND:
         switch (LOWORD(wParam)) {
            case IDOK :
               EndDialog(hDlg, LOWORD(wParam));
               return TRUE;

            case IDCANCEL :
               EndDialog(hDlg, LOWORD(wParam));
               return FALSE;

            default :
               return FALSE;
         }
   }
   return FALSE;
}

//! \brief Update BDM Firmware
//!
//! The BDM interface is reopened afterwards
//!
//! @param flashImage - The image to program into the BDM flash
//!
static int updateFirware(FlashImage *flashImage) {
   U32u version;
   int rc;
   int numDevices;
   int retry;
   U8  icpSoftwareVersion;
   U8  icpHardwareVersion;

   print("upDateFlash: Doing ICP reboot\r\n");
   // Cause BDM to reboot into ICP mode (ignored if already in ICP mode)
   _usbdm_setBoot();
   // Connection is lost so close it.
   bdm_usb_close();        // Connection is lost on reboot so close it.

   retry = 10;
   do { // Retry connection to allow for reboot, re-enumeration time
      Sleep(1000);
      numDevices=_opensourcebdm_init();
   } while ((numDevices <= 0) && (retry-- >0));

   rc = _opensourcebdm_open(0);
   if (rc != 0) {
      print("upDateFlash: Failed to re-open after reboot\r\n");
      return 1;
      }

   // Check version & in ICP mode
   version.longword    = _opensourcebdm_get_version();
   icpSoftwareVersion  = version.bytes[0]; // Software version from ICP code
   icpHardwareVersion  = version.bytes[1]; // Hardware version from ICP code

   print("upDateFlash: Checking in ICP mode\r\n");
   if ((icpSoftwareVersion != 0xFF) ||                     // Not in ICP mode or
       (icpHardwareVersion != availableHardwareVersion)) { // HW version incorrect
      print("upDateFlash: Failed to reboot to ICP mode or ICP HW incorrect\r\n");
      return 2;
   }

   // Erase Flash image area (ICP area is protected)
   print("upDateFlash: Erasing\r\n");
   rc = _usbdm_icp_erase( flashImage->loadAddr,   flashImage->size );
   if (rc != 0) {
      print("upDateFlash: Failed to erase\r\n");
      return 3;
      }
   // Program the new code
   print("upDateFlash: Programming\r\n");
   rc = _usbdm_icp_program( flashImage->loadAddr, flashImage->size, flashImage->data );
   if (rc != 0) {
      print("upDateFlash: Failed to program\r\n");
      return 4;
      }
   // Verify the new code
   print("upDateFlash: Verifying\r\n");
   rc = _usbdm_icp_verify(  flashImage->loadAddr, flashImage->size, flashImage->data );
   if (rc != 0) {
      print("upDateFlash: Failed to verify\r\n");
      return 5;
      }

   print("upDateFlash: Doing Normal reboot (ignore errors)\r\n");
   _usbdm_icp_reboot();    // Cause BDM to reboot into normal mode
   bdm_usb_close();        // Connection is lost on reboot so close it.
   retry = 10;
   do { // Retry connection to allow for reboot, re-enumeration time
      Sleep(1000);
      numDevices=_opensourcebdm_init();
   } while ((numDevices <= 0) && (retry-- >0));

   if ((numDevices <= 0) || (_opensourcebdm_open(0) != 0)) {
      print("upDateFlash: Failed to re-open after 2nd reboot\r\n");
      return 6;
      }
   print("upDateFlash: Successful\r\n");
   return 0;
}

//! \brief Display Firmware update Dialogue & optionally updates BDM Firmware
//!
//! The BDM interface is reopened
//!
//! @param hDLLInst     DLL Handle (to obtain resources)
//!
void checkFirmware(HINSTANCE hDLLInst){
int rc;
static int recurse = 0;
U32u version;
FlashImage *flashImage;

   print("checkFirmware: hDLLInst = %d\r\n", hDLLInst);

   // Check if already updating Flash
   if (recurse) {
      print("checkFirmware: recursing\r\n", hDLLInst);
      return;
      }
   recurse = 1;

   version.longword        = _opensourcebdm_get_version();
   currentSoftwareVersion  = version.bytes[0]; // Software version from BDM code
   currentHardwareVersion  = version.bytes[1]; // Hardware version from BDM code

   print("checkFirmware: version = %4.4X\r\n", version);

   if ((currentHardwareVersion & 0xC0) == 0) // ICP not available via driver unless JMxx, UFxx
      return;

   print("checkFirmware: Current BDM version = %1X.%1X, Hardware version = %2X\r\n",
         (currentSoftwareVersion>>4)&0x0F,currentSoftwareVersion&0x0F,
         currentHardwareVersion);

   // Check each available flash image
   for (flashImage=flashImages;
        flashImage<flashImages+sizeof(flashImages)/sizeof(flashImages[0]);
        flashImage++) {
      availableSoftwareVersion = flashImage->softwareVersion;
      availableHardwareVersion = flashImage->hardwareVersion;

      print("checkFirmware: New Flash version   = %1X.%1X, Hardware version = %2X\r\n",
            (availableSoftwareVersion>>4)&0x0F,availableSoftwareVersion&0x0F,
             availableHardwareVersion );
      if ((currentHardwareVersion==availableHardwareVersion) && // Correct H/W version and
          ((currentSoftwareVersion<availableSoftwareVersion) || // later S/W version or
           (currentSoftwareVersion == 0xFF)) ) {                // already in ICP mode
         break;
      }
   }
   if (flashImage>=flashImages+sizeof(flashImages)/sizeof(flashImages[0])) {
      // No update found
      print("checkFirmware: No update available \r\n");
      return;
   }
   rc = DialogBox(hDLLInst, (LPCTSTR)IDD_FLASH_UPDATE, 0, (DLGPROC)Configure);

   if (rc <0) { // Dialogue failure
      LPVOID lpMsgBuf;
      DWORD dw = GetLastError();

      print("checkFirmware: rc = %d\r\n", rc);

      FormatMessage(
          FORMAT_MESSAGE_ALLOCATE_BUFFER|FORMAT_MESSAGE_FROM_SYSTEM|FORMAT_MESSAGE_IGNORE_INSERTS,
          NULL,
          dw,
          MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
          (LPTSTR) &lpMsgBuf,
          0, NULL );
      print("checkFirmware: Dialogue Error %s", lpMsgBuf);
   }
   else if (rc == TRUE) {
      rc = updateFirware(flashImage);
      if (rc != 0) {
         // Do something to alert user!
      }
   }

   recurse = 0;
}



