/*! \file
    \brief Entry points for tbdml.dll library

    \verbatim
    Turbo BDM Light - interface DLL
    Based on TBDML interface DLL by
    Copyright (C) 2005  Daniel Malik

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    \endverbatim

*/
#include <windows.h>
#include <stdio.h>
#include "tbdml.h"
#include "Configure.h"
#include "log.h"
#include "version.h"
#include "ICP.h"

//#define TBDML_API  // Mark this file as exporting TBDML functions

#include "common_usb.h"
#include "commands.h"
#include "Common.h"

/*
 *  Define symbol "LOG" in "log.h" during compilation to produce a log file tbdml_dll.log
 */

static unsigned char usb_data[MAX_PACKET_SIZE+1]; // Extra +1 for debug byte
static HINSTANCE hDLLInst;
//static int initialised = 0;

//! Get version of the DLL in BCD format
//!
//! @return 8-bit version number (2 BCD digits, major-minor)
//!
TBDML_API unsigned char  _tbdml_dll_version(void) {
	return(TBDML_DLL_VERSION);
}

//! Initialises USB interface
//!
//! This must be done before any device may be opened.
//!
//! @return Number of BDM devices found
//!
TBDML_API unsigned char _tbdml_init(void) {
	unsigned char i;
	bdm_usb_find_devices();	/* look for devices on all USB busses */
	i=bdm_usb_cnt();
	print("OSBDM_INIT: Usb initialised, found %d device(s)\r\n",i);
	return (i);					/* count of devices found */
}

//! Opens a device
//!
//! @param device_no Number (0..N) of device to open.
//! A device must be open before any communication with the device can take place.
//! @return 0 => Success,\n !=0 => Fail
//!
TBDML_API unsigned char _tbdml_open(unsigned char device_no) {
int rc = 0;

   print("OSBDM_OPEN: Trying to open device #%d\r\n",device_no);
   rc = bdm_usb_open(device_no);
   print("OSBDM_OPEN: rc = #%d\r\n", rc);
   if ((rc == 0) & (bdm_usb_cnt() == 1)) {
     // Successful open and only one device available
     checkFirmware(hDLLInst); // Check Firmware version
   }
   return (rc);
}

//! Closes currently open device
//!
TBDML_API void _tbdml_close(void) {
	print("OSBDM_CLOSE: Trying to close the device\r\n");

	// Tell the BDM that we're finished (power down)
   usb_data[0]=2;  /* receive 2 byte */
   usb_data[1]=CMD_SET_TARGET;
   usb_data[2]=0xFF;
   bdm_usb_recv_ep0(usb_data);
	bdm_usb_close();
}

//! Gets version of the interface (HW and SW) in BCD format
//!
//! @return 32-bit version number (4 bytes)
//!  LSB = software
//!      = hardware
//!      = ICP version
//!  MSB = ICP type - 0x00 for JB16, 0x80 for JM60
TBDML_API unsigned int  _tbdml_get_version(void) {
   usb_data[0] = 5;  // receive up to 3 bytes
	usb_data[1] = CMD_GET_VER;
	bdm_usb_recv_ep0(usb_data);
	return (*((U32 *)(usb_data+1)));
}

//! Get status of the last command (either the command number or CMD_FAILED or CMD_UNKNOWN)
//!
//! @return 0 => Success,\n 1 => Failure
//!
TBDML_API unsigned char _tbdml_get_last_sts(void) {
   usb_data[0] = 1;  // receive up to 1 byte
	usb_data[1] = CMD_GET_LAST_STATUS;
	bdm_usb_recv_ep0(usb_data);
	return ((usb_data[0]==CMD_FAILED)||(usb_data[0]==CMD_UNKNOWN));
}

//! Sets target MCU type
//!
//! @param target_type type of target, see \ref target_type_e
//!
//! @return 0 => Success,\n !=0 => Fail
//!
TBDML_API unsigned char _tbdml_set_target_type(target_type_e target_type) {

   configureDialog(hDLLInst, target_type); // Do configuration dialogue

   usb_data[0] = 1;  // receive up to 1 byte
	usb_data[1] = CMD_SET_TARGET;
	usb_data[2] = target_type;
	bdm_usb_recv_ep0(usb_data);
	return (usb_data[0]!=CMD_SET_TARGET);
}


//! Connects to Target
//!
//! @return 0 => Success,\n !=0 => Fail
//!
TBDML_API unsigned char _tbdml_target_sync(void) {
   usb_data[0] = 1;  // receive up to 1 byte
	usb_data[1] = CMD_CONNECT;
	bdm_usb_recv_ep0(usb_data);
	return (usb_data[0]!=CMD_CONNECT);
}

//! Resets the target to normal or special mode
//!
//! @param target_mode see \ref target_mode_e
//!
//! @return 0 => Success,\n !=0 => Fail
//!
TBDML_API unsigned char _tbdml_target_reset(target_mode_e target_mode) {
   usb_data[0] = 1;  // receive up to 1 byte
	usb_data[1] = CMD_RESET;
	usb_data[2] = target_mode;
	bdm_usb_recv_ep0(usb_data);
	return (usb_data[0]!=CMD_RESET);
}

//! Fills user supplied structure with state of BDM communication channel
//!
//! @param bdm_status Pointer to structure to receive status, see \ref bdm_status_t
//!
//! @return 0 => Success,\n !=0 => Fail
//!
TBDML_API unsigned char _tbdml_bdm_sts(bdm_status_t *bdm_status) {
   unsigned char BDMStatus;
   static const char *connectionStates[] = {"Not Connected", "Sync Speed",
                                            "Guess Speed", "Manual Speed"};
   usb_data[0] = 2;  // receive up to 2 bytes
	usb_data[1] = CMD_GET_STATUS;
	bdm_usb_recv_ep0(usb_data);
	if (usb_data[0]!=CMD_GET_STATUS)
	   return (1);
	BDMStatus = usb_data[1];
	bdm_status->ackn_state       = (ackn_e)  (BDMStatus    &0x01);
	bdm_status->reset_state      = (reset_e)((BDMStatus>>1)&0x01);
	bdm_status->connection_state = (speed_e)((BDMStatus>>3)&0x03);

	print("Status = %s, %s, %s, %s\r\n", bdm_status->ackn_state?"Ackn":"Wait",
	                                     connectionStates[bdm_status->connection_state],
	                                     bdm_status->reset_state?"Reset":"No Reset",
	                                     (BDMStatus&0x04)?"Power":"No Power" );
	return(0);
}


//! Starts target execution from current PC address
//!
//! @return 0 => Success,\n !=0 => Fail
//!
TBDML_API unsigned char _tbdml_target_go(void) {
   usb_data[0] = 1;  // receive up to 1 byte
	usb_data[1]=CMD_GO1;
	bdm_usb_recv_ep0(usb_data);
	return(usb_data[0]!=CMD_GO1);
}

//! Steps over a single target instruction
//!
//! @return 0 => Success,\n !=0 => Fail
//!
TBDML_API unsigned char _tbdml_target_step(void) {
   usb_data[0] = 1;  // receive up to 1 byte
	usb_data[1]=CMD_STEP1;
	bdm_usb_recv_ep0(usb_data);
	return(usb_data[0]!=CMD_STEP1);
}

//! Brings the target into active background mode.  The target will be halted.
//!
//! @return 0 => Success,\n !=0 => Fail
//!
TBDML_API unsigned char _tbdml_target_halt(void) {
   usb_data[0] = 1;  // receive up to 1 byte
	usb_data[1]=CMD_HALT;
	bdm_usb_recv_ep0(usb_data);
	return(usb_data[0]!=CMD_HALT);
}

//! Determine crystal frequency of the target in MHz
//!
//! @return Float representing the Crystal Frequency of target
//!
TBDML_API float _tbdml_get_speed(void) {
   usb_data[0] = 3;  // receive up to 3 bytes
	usb_data[1] = CMD_READ_SPEED1;
	bdm_usb_recv_ep0(usb_data);
	return (60.0 * 2 * 128 / (*(U16 *)(usb_data+1)));
}

//! Sets the communication speed
//!
//! @param crystal_frequency crystal frequency in MHz
//! This is used to determine the BDM communication speed (assumed
//! to be half the crystal frequency given).
//!
//! @return 0 => Success,\n !=0 => Fail
//!
TBDML_API unsigned char _tbdml_set_speed(float crystal_frequency) {
	/* the command parameter is the expected length
	 *  of SYNC pulse expressed in 60MHz ticks */
   usb_data[0] = 1;  // receive up to 1 byte
	usb_data[1] = CMD_SET_SPEED1;
	*(U16 *)(usb_data+2) = 60.0 * 2 * 128 / crystal_frequency;
	bdm_usb_recv_ep0(usb_data);
	return (usb_data[0]!=CMD_SET_SPEED1);
}

//===========================================================

//! Reads one byte from memory
//!
//! @param address 16-bit memory address
//!
//! @return 8-bit value
//!
TBDML_API unsigned char _tbdml_read_byte(unsigned int address) {
   usb_data[0]          = 2;  // receive up to 2 bytes
	usb_data[1]          = CMD_READ_8;
	*(U16 *)(usb_data+2) = address;
	bdm_usb_recv_ep0(usb_data);
	return (usb_data[1]);
}

//! Writes one byte to memory
//!
//! @param address 16-bit memory address
//! @param data    8-bit data value
//!
TBDML_API void _tbdml_write_byte(unsigned int address, unsigned char data) {
   usb_data[0]          = 1;  // receive up to 1 byte
	usb_data[1]          = CMD_WRITE_8;
	*(U16 *)(usb_data+2) = address;
	usb_data[4]          = data;
	bdm_usb_recv_ep0(usb_data);
}

//! Reads block from memory
//!
//! @param addr    16-bit memory address
//! @param count   8-bit byte count
//! @param data    Pointer to buffer to contain data
//!
TBDML_API void _tbdml_read_block(unsigned int addr,
                                         unsigned int count,
                                         unsigned char *data) {

   unsigned int blockSize;
   // Make multiple of 2, Allow for status[0]+data[1...]
   const unsigned int MaxDataSize = (MAX_PACKET_SIZE-1) & ~1;

   while (count > 0) {
      blockSize = count;
      if (blockSize > MaxDataSize)
         blockSize = MaxDataSize;

      /* if address is odd and we are going to read more than one
       *  block make the first packet 1 byte short to align the address */
      if ((count>MaxDataSize) && (addr&0x1))
         blockSize--;

      usb_data[1]             = CMD_READ_BLOCK1;
      *((U16 *)(usb_data+2))  = addr;
      usb_data[4]             = blockSize;
      #ifdef FAST_BLOCK
         usb_data[0] = 4;
         tbdml_usb_send_ep2(usb_data);
         usb_data[0] = blockSize+1;
         tbdml_usb_recv_ep2(usb_data);
      #else
         usb_data[0] = blockSize+1;
         bdm_usb_recv_ep0(usb_data);
      #endif
      memcpy(data, usb_data+1, blockSize);

      data  += blockSize;   // update location in buffer
      addr  += blockSize;   // update memory address
      count -= blockSize;   // update count
   }
}

//! Writes block to memory
//!
//! @param addr    16-bit memory address
//! @param count   8-bit byte count
//! @param data    Pointer to buffer containing data
//!
TBDML_API void _tbdml_write_block(unsigned int addr,
                                          unsigned int count,
                                          unsigned char *data) {
unsigned retry;

   unsigned int blockSize;
   // Make multiple of 2, allow for status[0]+setup[1..4] + data[5...]
   const unsigned int MaxDataSize = (MAX_PACKET_SIZE-5)&~1;

   while (count > 0) {
      blockSize = count;
      if (blockSize > MaxDataSize)
         blockSize = MaxDataSize;

      /* if address is odd and we are going to read more than one
       *  block make the first packet 1 byte short to align the address */
      if ((count>MaxDataSize)&&(addr&0x1))
         blockSize--;

      usb_data[0]             = blockSize+4;
      usb_data[1]             = CMD_WRITE_BLOCK1;
      *((U16 *)(usb_data+2))  = addr;
      usb_data[4]             = blockSize;
      memcpy(usb_data+5, data, blockSize);

      #ifdef FAST_BLOCK
         tbdml_usb_send_ep2(usb_data); /* Tx data */
         usb_data[0]=1;                /* read status */
         tbdml_usb_recv_ep2(usb_data);
      #else
         // ToDo - error check
         bdm_usb_send_ep0(usb_data);
         // do handshake
         retry = 10;
         do {
            usb_data[0]=1;  /* receive 1 byte */
            usb_data[1]=CMD_GET_LAST_STATUS;
            if (bdm_usb_recv_ep0(usb_data) == 0)
                break;
            Sleep(50 /* ms */); // So we don't tie up the USB
         } while (retry-- > 0);
      #endif

      data  += blockSize;   // update location in buffer
      addr  += blockSize;   // update address
      count -= blockSize;   // update count
   }
}

//! Reads one word from memory (address must be aligned)
//!
//! @param address 16-bit memory address
//!
//! @return 16-bit data value
//!
TBDML_API unsigned int  _tbdml_read_word(unsigned int address) {
   usb_data[0] = 3;  // receive up to 3 bytes
	usb_data[1] = CMD_READ_16;
	*((U16 *)(usb_data+2)) = address;
	bdm_usb_recv_ep0(usb_data);
	return (*(U16 *)(usb_data+1));
}

//! (HCS12) Writes one word to memory (address must be aligned)
//!
//! @param address 16-bit memory address
//! @param data 16-bit data value
//!
TBDML_API void  _tbdml_write_word(unsigned int address, unsigned int data) {
   usb_data[0] = 1;  // receive up to 1 byte
	usb_data[1] = CMD_WRITE_16;
	*(U16 *)(usb_data+2) = address;
	*(U16 *)(usb_data+4) = data;
	bdm_usb_recv_ep0(usb_data);
}

//! (HCS12) Reads one byte from the BDM address space
//!
//! @param address 16-bit memory address
//! @return 8-bit data value
//!
#define HC12_BDMSTS     0xFF01   //!< Address of HC12 BDM Status register

TBDML_API unsigned char _tbdml_read_bd(unsigned int address) {
   unsigned char BDMStatus;
   usb_data[0] = 2;  // receive up to 2 bytes
	usb_data[1] = CMD_READ_BD;
	*((U16 *)(usb_data+2)) = address;
	bdm_usb_recv_ep0(usb_data);
	BDMStatus = usb_data[1];
   if (address == HC12_BDMSTS) {
      print("Status = %s, %s, %s\r\n", (BDMStatus&0x80)?"ENBDM":"-",
                                       (BDMStatus&0x40)?"BDACT":"-",
                                       (BDMStatus&0x04)?"CLKSW":"-" );
	}
	return (usb_data[1]);
}

//! (HCS12) Writes one byte to the BDM address space
//!
//! @param address 16-bit memory address
//! @param data    8-bit data value
//!
//! @return 0 => Success,\n !=0 => Fail
//!
TBDML_API unsigned char _tbdml_write_bd(unsigned int address, unsigned char data) {
   usb_data[0] = 2;  // receive up to 2 bytes
	usb_data[1] = CMD_WRITE_BD;
	*((U16 *)(usb_data+2)) = address;
	usb_data[4] = data;
	bdm_usb_recv_ep0(usb_data);
	return (usb_data[0]!=CMD_WRITE_BD);
}

//! (HCS12, HCS08 & RS08) Reads contents of core registers
//!
//! @param  registers   pointer to structure to receive the register values
//!
//! @return 0 => Success,\n !=0 => Fail
//!
TBDML_API unsigned char _tbdml_read_regs(registers_t *registers) {
   usb_data[0] = 13;  // receive up to 13 bytes
	usb_data[1]=CMD_READ_REGS;
	bdm_usb_recv_ep0(usb_data);
   // Note: HCS08/RS08 structs are compatible
	registers->hc12.pc = (*((U16*)(usb_data+1)));
	registers->hc12.sp = (*((U16*)(usb_data+3)));
	registers->hc12.ix = (*((U16*)(usb_data+5)));
	registers->hc12.iy = (*((U16*)(usb_data+7)));
	registers->hc12.d  = (*((U16*)(usb_data+9)));
	registers->hc12.ccr= usb_data[11];
	return(usb_data[0]!=CMD_READ_REGS);
}

//! (HCS12, HCS08 & RS08) Writes PC (HCS12, HC08), CCR+PC (RS08)
//!
//! @param value 16-bit value
//!
TBDML_API void _tbdml_write_reg_pc(unsigned int value) {
   usb_data[0] = 1;  // receive up to 1 bytes
	usb_data[1] = CMD_WRITE_REG_PC;
	*(U16 *)(usb_data+2) = value;
	bdm_usb_recv_ep0(usb_data);
}

//! (HCS12, HCS08 & RS08) Writes SP (HCS12, HCS08), Shadow PC (RS08)
//!
//! @param value 16-bit value
//!
TBDML_API void _tbdml_write_reg_sp(unsigned int value) {
   usb_data[0] = 1;  // receive up to 1 byte
	usb_data[1] = CMD_WRITE_REG_SP;
	*(U16 *)(usb_data+2) = value;
	bdm_usb_recv_ep0(usb_data);
}

//! (HCS12 & HCS08) Writes IX (HCS12), H:X (HCS08)
//!
//! @param value 16-bit value
//!
TBDML_API void _tbdml_write_reg_x(unsigned int value) {
   usb_data[0] = 1;  // receive up to 1 byte
	usb_data[1] = CMD_WRITE_REG_X;
	*(U16 *)(usb_data+2) = value;
	bdm_usb_recv_ep0(usb_data);
}

//! (HCS12) Writes IY
//!
//! @param value 16-bit value
//!
TBDML_API void _tbdml_write_reg_y(unsigned int value) {
   usb_data[0] = 1;  // receive up to 1 byte
	usb_data[1] = CMD_WRITE_REG_Y;
	*(U16 *)(usb_data+2) = value;
	bdm_usb_recv_ep0(usb_data);
}

//! (HCS12, HCS08 & RS08) Writes D=B:A (HC12) or A (HCS08)
//!
//! @param value 16-bit / 8-bit value
//!
TBDML_API void _tbdml_write_reg_d(unsigned int value) {
   usb_data[0] = 1;  // receive up to 1 byte
	usb_data[1] = CMD_WRITE_REG_D;
	*(U16 *)(usb_data+2) = value;
	bdm_usb_recv_ep0(usb_data);
}

//! (HCS12 & HCS08) Writes CCR
//!
//! @param value 8-bit value
//!
TBDML_API void _tbdml_write_reg_ccr(unsigned int value) {
   usb_data[0] = 1;  // receive up to 1 byte
	usb_data[1] = CMD_WRITE_REG_CCR;
	*(U16 *)(usb_data+2) = value;
	bdm_usb_recv_ep0(usb_data);
}

//! Read target BDM status register (BDCSC, BDMSTS or XCSR as appropriate)
//!
//! @return 8-bit value
//!
//!
TBDML_API unsigned char _tbdml_read_status(void){
   unsigned char BDMStatus;
   usb_data[0] = 2;  // receive up to 2 bytes
   usb_data[1]=CMD_READ_STATUS;
   bdm_usb_recv_ep0(usb_data);
	BDMStatus = usb_data[1];
   print("Status = %s, %s, %s, %s\r\n", (BDMStatus&0x80)?"ENBDM":"-",
                                        (BDMStatus&0x40)?"BDACT":"-",
                                        (BDMStatus&0x20)?"BKPTEN":"-",
                                        (BDMStatus&0x08)?"CLKSW":"-" );
   return(usb_data[1]);
}

//! Write to BDM Control Register (BDCSC, BDMSTS or XCSR as appropriate)
//!
//! @param value 8-bit value
//!
TBDML_API void _tbdml_write_control(unsigned char value){
   usb_data[0] = 1;  // receive up to 1 byte
   usb_data[1] = CMD_WRITE_CONTROL;
   usb_data[2] = value;
   bdm_usb_recv_ep0(usb_data);
}

//! (HCS08 & RS08) Write to Breakpoint Register
//!
//! @param value 16-bit value
//!
TBDML_API void _tbdml_write_bkpt(unsigned int value){
   usb_data[0] = 1;  // receive up to 1 byte
   usb_data[1] = CMD_WRITE_BKPT;
   *(U16 *)(usb_data+2) = (U16)value;
   bdm_usb_recv_ep0(usb_data);
}

//! (HCS08 & RS08) Read from Breakpoint Register
//!
//! @return 16-bit value
//!
TBDML_API unsigned int  _tbdml_read_bkpt(void){
   usb_data[0] = 3;  // receive up to 3 bytes
   usb_data[1] = CMD_READ_BKPT;
   bdm_usb_recv_ep0(usb_data);
   return (*(U16*)(usb_data+1));
}

/*! Set BDM for ICP mode & reboot
 *
 *  The BDM resets in ICP mode after this command
 *  @note Communication is lost.
 */
TBDML_API void _usbdm_setBoot(void){

   usb_data[0] = 3;  // receive up to 3 bytes
   usb_data[1] = CMD_SET_BOOT;
   usb_data[2] = 'B';
   usb_data[3] = 'O';
   usb_data[4] = 'O';
   usb_data[5] = 'T';

   bdm_usb_send_ep0(usb_data);
}

//! ICP mode - reboot
//!
//! The BDM does a normal reset
//!
//! Usually used in ICP mode to reset to normal (BDM) mode.
//! @note Communication is lost.  \n
//!       This will do the usual flash validity checks before entering BDM mode.
//!
TBDML_API void _usbdm_icp_reboot( void ) {

   bdm_usb_raw_send_ep0(ICP_REBOOT,
                        0,
                        0,
                        0,
                        NULL);
}


//! ICP mode - program BDM Flash memory
//!
//! @param addr    16-bit memory address
//! @param count   8-bit byte count
//! @param data    Pointer to buffer containing data
//!
TBDML_API unsigned int _usbdm_icp_program(unsigned int  addr,
                                                  unsigned int  count,
                                                  unsigned char *data) {

   unsigned int  dataSize;     // Bytes to program this row
   unsigned char rc;
   unsigned int  temp;
            int  doBlock;

   // Have to program within a row in each iteration
   // so there are size and alignment issues.
   while (count > 0) {
      // Bytes to end of current row
      dataSize = ICP_JM60_ROWSIZE - (addr & (ICP_JM60_ROWSIZE-1));
      if (dataSize > count)  // Take down to number actually remaining
         dataSize = count;

      // Check if empty block (all 0xFF)
      doBlock = FALSE;
      for (temp = 0; temp<dataSize; temp++)
         doBlock = doBlock || (data[temp] != 0x00);

      if (doBlock) { // Only program non-Blank blocks
         rc = bdm_usb_raw_send_ep0(ICP_PROGRAM_ROW,
                                   addr,
                                   0,
                                   dataSize,
                                   data);
         if (rc != ICP_RC_OK)
            return rc;

         rc = _usbdm_icp_get_result();

         if (rc != ICP_RC_OK)
            return rc;
      }
      data  += dataSize;   // update location in buffer
      addr  += dataSize;   // update address
      count -= dataSize;   // update count
   }
   return ICP_RC_OK;
}

//! ICP mode - verify BDM Flash memory
//!
//! @param addr    16-bit memory address
//! @param count   8-bit byte count
//! @param data    Pointer to buffer containing data
//!
TBDML_API unsigned int _usbdm_icp_verify(unsigned int  addr,
                                         unsigned int  count,
                                         unsigned char *data) {
   unsigned       int  dataSize;
   unsigned       char rc;

   // Can verify MaxDataSize bytes at a time
   // Doesn't have alignment issues
   while (count > 0) {
      dataSize = count;
      if (dataSize > ICP_JM60_ROWSIZE)
         dataSize = ICP_JM60_ROWSIZE;

      rc = bdm_usb_raw_send_ep0(ICP_VERIFY_ROW,
                                addr,
                                0,
                                dataSize,
                                data);
      if (rc != ICP_RC_OK)
         return rc;

      rc = _usbdm_icp_get_result();

      if (rc != ICP_RC_OK)
         return rc;

      data  += dataSize;   // update location in buffer
      addr  += dataSize;   // update address
      count -= dataSize;   // update count
   }
   return ICP_RC_OK;
}

//! ICP mode - erase BDM Flash memory
//!
//! @param addr    16-bit memory address
//! @param count   8-bit byte count
//!
TBDML_API unsigned int _usbdm_icp_erase(unsigned int addr,
                                                unsigned int count) {
   unsigned int blockSize;
   unsigned char rc;

   // Flash memory is organised in blocks
   // Only an entire block can be erased at a time
   while (count > 0) {
      blockSize = count;
      if (blockSize > ICP_JM60_BLOCKSIZE)
         blockSize = ICP_JM60_BLOCKSIZE;

      rc = bdm_usb_raw_send_ep0(ICP_ERASE_PAGE,
                                addr,
                                0,
                                0,
                                NULL);
      if (rc != ICP_RC_OK)
         return rc;

      rc = _usbdm_icp_get_result();

      if (rc != ICP_RC_OK)
         return rc;

      addr  += blockSize;   // update address
      count -= blockSize;   // update count
   }
   return ICP_RC_OK;
}

//! ICP mode - get result of last ICP command
//!
TBDML_API unsigned int _usbdm_icp_get_result( void ) {

   unsigned char rc;
   unsigned char buff[10];

   rc = bdm_usb_raw_recv_ep0( ICP_GET_RESULT,
                              0,
                              0,
                              sizeof(buff),
                              buff);
   return buff[0];
}

#if defined (COMMANDLINE)
//! Execute debug command (various, see \ref DebugSubCommands)
//!
//! @return 16-bit value
//!
void _opensourcebdm_debug(char *usb_data) {
   usb_data[0] = 7;  // receive up to 7 bytes
   usb_data[1] = CMD_DEBUG;
   bdm_usb_recv_ep0(usb_data);
}
#endif

static
//! Initialise DLL
//!
void _tbdml_initialize(HINSTANCE hDLLInst){

   openLogFile(NULL);

   bdm_usb_init();       /* initialise USB */

//   initialised = 1;
}

static
//! Un-initialise DLL
//!
void _tbdml_uninitialize(void) {
   closeLogFile();
//   initialised = 0;
}

#if defined(DLL)
//!
BOOL TBDML_API DllMain(HINSTANCE _hDLLInst,
                       DWORD fdwReason, LPVOID lpvReserved) {
   hDLLInst = _hDLLInst;
   switch (fdwReason) {
      case DLL_PROCESS_ATTACH:
         _tbdml_initialize(hDLLInst);
         break;
      case DLL_PROCESS_DETACH:
         _tbdml_uninitialize();
         break;
      case DLL_THREAD_ATTACH:
         break;
      case DLL_THREAD_DETACH:
         break;
 }
 return TRUE;
}
#endif
