/*! \file
    \brief Open Source BDM - BDM communication.
    
    Contains low level BDM interface routines.

   \verbatim
   *** Prominent Notice - This software was modified from TBDML software - 12/05 ***
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
   \endverbatim

   \verbatim
   Change Hisory
   -=========================================================================
   | 11 Nov 2008 | Unified Timer MACROS                                 - pgo
   |  5 Nov 2008 | Re-arrangement of Target Vdd control routines        - pgo
   |  1 Nov 2008 | Minor mods & corrected comments in bdm_syncMeasure() - pgo
   |  1 Nov 2008 | Removed glitch in bdm_syncMeasure()                  - pgo
   | 23 Oct 2008 | Reviewed stack size                                  - pgo
   | 22 Oct 2008 | Increased bdm_status.waitX size                      - pgo
   | 22 Oct 2008 | Changed RESET macros - more consistent               - pgo
   |  4 Oct 2008 | Renamed some MACROs for consistency                  - pgo
   | 29 Aug 2008 | Disabled ACMP0 as it clashes with Vdd5_En* on CLC    - pgo
   | 10 Aug 2008 | Fixed bug in multi-line MACROs                       - pgo
   | 10 Jul 2008 | Modified Speed Guessing routine (now smarter!)       - pgo
   | 10 Jul 2008 | Modified Reset code slightly                         - pgo
   |  9 Jul 2008 | Modified Rx routine mask to remove glitch            - pgo
   |  7 Jul 2008 | Modified Reset sequence to properly Cycle Vdd        - pgo
   |  2 Jul 2008 | More thorough checking for target Vdd                - pgo
   |  2 Jul 2008 | Modified using Alt BDM clock handling                - pgo
   | 19 Jun 2008 | Expanded Flash programming options of RS08           - pgo
   | 14 Jun 2008 | Added RS08 Trim value calculation                    - pgo
   | 10 Jun 2008 | Added Colfire V1 interface                           - pgo
   |  6 May 2008 | Ported to JB16 from JM60 - lotsa changes             - pgo
   | 12 Apr 2008 | Added reconnect before software reset                - pgo
   |  7 Apr 2008 | Fixed WAIT_WITH_TIMEOUT.. macros                     - pgo
   |  7 Apr 2008 | Fixed glitch in bdm_init                             - pgo
   |  6 Apr 2008 | Added Vdd change detect handler & POR                - pgo
   | 23 Mar 2008 | Fixed several BMD commands not using ACKN            - pgo
   |  3 Mar 2008 | Started changes for JM60 - lots                      - pgo
   +=========================================================================
   \endverbatim
*/

#include <hidef.h>          /* common defines and macros */
#include "derivative.h"
#include "Common.h"
#include "Configure.h"
#include "bdm.h"
#include "bdmMacros.h"
#include "commands.h"
#include "usb.h"
#include "TargetDefines.h"
#include "CmdProcessing.h"

//=================================================================================
#define BDM_SYNC_REQ      350U // us - length of the longest possible SYNC REQUEST pulse (128 BDM cycles @ 400kHz = 320us plus some extra time) 
#define SYNC_TIMEOUT      460U // us - longest time for the target to completed a SYNC pulse (16+128+margin cycles @ 400kHz = 375us) 
#define ACKN_TIMEOUT      375U // us - longest time after which the target should produce ACKN pulse (150 cycles @ 400kHz = 375us) 
#define SOFT_RESET       1280U // us - longest time needed for soft reset of the BDM interface (512 BDM cycles @ 400kHz = 1280us)
#define RESET_SETTLE     3000U // us - time to wait for signals to settle in us, this should certainly be longer than the soft reset time
#define BKGD_WAIT        2000U // us - time to hold BKGD pin low after reset pin rise for special modes (allowance made for slow Reset rise)
#define VDD_RISE_TIME    2000U // us - minimum time to allow for controlled target Vdd rise
#define RESET_LENGTH      100U // ms - time of RESET assertion
#define RESET_WAIT        270U // ms - how long to wait for the reset pin to come high
#define RESET_RECOVERY      5U // ms - how long to wait after reset before new commands are allowed


/* Function prototypes */
U8          bdm_syncMeasure(void);
static void bdm_idle(void);
static U8   bdmHC12_alt_speed_detect(void);
static void bdm_checkTargetVdd(void);

static const bdm_status_t bdm_statusDefault = {T_OFF,WAIT,NO_RESET_ACTIVITY,NO_INFO,0,0,0,1};

//========================================================
//
#pragma DATA_SEG __SHORT_SEG Z_PAGE
// MUST be placed into the direct segment (assumed in ASM code).
static U8 bitCount; //!< Used as a general purpose variable in the bdm_Tx() & bdm_Rx() etc.
static U8 rxTiming1; //!< bdm_Rx timing constant #1
static U8 rxTiming2; //!< bdm_Rx timing constant #2
static U8 rxTiming3; //!< bdm_Rx timing constant #3
static U8 txTiming1; //!< bdm_Tx timing constant #1
static U8 txTiming2; //!< bdm_Tx timing constant #2
static U8 txTiming3; //!< bdm_Tx timing constant #3

// pointers to current bdm_Rx & bdm_Tx routines
U8   (*bdm_rx_ptr)(void) = bdm_rxEmpty; //!< pointers to current bdm_Rx routine
void (*bdm_tx_ptr)(U8)   = bdm_txEmpty; //!< pointers to current bdm_Tx routine

//! Status of the BDM 
//!
//! see \ref bdm_status_t
bdm_status_t bdm_status   = {0}; // Status of the BDM interface

//! Options for the BDM
//!
//! see \ref bdm_option_t
bdm_option_t bdm_option   = {0,BDM_TARGET_VDD_OFF,FALSE,FALSE,TRUE,TRUE,FALSE,DEFAULT,FALSE,0};

//! Trim value read from RS08 target Flash or calculated from desired Target clock speed
volatile U8 savedTrimValue = 0xFF;

//! State of RS08 Flash programming,  see \ref flashState_t
volatile U8 flashState     = FLASH_IDLE;

//========================================================
//
#pragma DATA_SEG DEFAULT

U8 flashProgrammingBuffer[5 + MAX_FLASH_PROGRAM]; // 5 byte header + data to program

//! Flash writing program for download into RS08KA1/KA2 Direct RAM
static const U8 RS08KA2_Flash_Program[48+3]= {
   48,                                  // Size of program to load
   RS08KA2_DIR_RAM>>8, RS08KA2_DIR_RAM, // Where to load program
   // Program code
   0x3E,0x08,0x1F,0x10,0xD1,0xCA,0xFF,0xC9,0xEF,0xEE,0xCB,0x4B,0xFE,0x3E,0x08,0x1F,
   0x16,0xD1,0xCC,0x4B,0xFE,0xCA,0xFF,0xC0,0xEF,0xCE,0x4E,0x09,0x0F,0xEE,0xCD,0x4B,
   0xFE,0x29,0x3B,0x00,0xF2,0x3E,0x08,0x1F,0x11,0xD1,0xCB,0x4B,0xFE,0x17,0xD1,0xBF, 
   };

//! Flash writing program for download into RS08KA4/KA8 Direct RAM
static const U8 RS08KA8_Flash_Program[48+3]= {
   48,                                  // Size of program to load
   RS08KA8_DIR_RAM>>8, RS08KA8_DIR_RAM, // Where to load program
   // Program code
   0x3E,0x08,0x1F,0x10,0xD1,0xCA,0xFF,0xC9,0xEF,0xEE,0xCB,0x4B,0xFE,0x3E,0x08,0x1F,
   0x16,0xD1,0xCC,0x4B,0xFE,0xCA,0xFF,0xC0,0xEF,0xCE,0x4E,0x09,0x0F,0xEE,0xCD,0x4B,
   0xFE,0x29,0x3B,0x00,0xF2,0x3E,0x08,0x1F,0x11,0xD1,0xCB,0x4B,0xFE,0x17,0xD1,0xBF,
   };

//! Flash writing program for download into RS08LA8 Direct RAM
static const U8 RS08LA8_Flash_Program[48+3]= {
   48,                                  // Size of program to load
   RS08LA8_DIR_RAM>>8, RS08LA8_DIR_RAM, // Where to load program
   // Program code
   0x3E,0x09,0x1F,0x10,0xD1,0xCA,0xFF,0xC9,0xEF,0xEE,0xCB,0x4B,0xFE,0x3E,0x09,0x1F,
   0x16,0xD1,0xCC,0x4B,0xFE,0xCA,0xFF,0xC0,0xEF,0xCE,0x4E,0x09,0x0F,0xEE,0xCD,0x4B,
   0xFE,0x29,0x3B,0x00,0xF2,0x3E,0x09,0x1F,0x11,0xD1,0xCB,0x4B,0xFE,0x17,0xD1,0xBF,
   };                                                                        

//==============================================================
//  Timer Usage:
//     TPM2     - General interval timer (no I/O pins)
//     TPM1-CHx - RESET_IN pin, Reset detection (IC falling edge detection)
//   TPM1-CHy - BDM_IN pin, SYNC measuring, ACKN detection (IC rising & falling edges)
//   TPM1-CH2 - Timeouts (Output compare)

//==============================================================*/

// TPM1 (SYNC & ACKN) timer constants, 24MHz ticks
#define ACKN_TIMER_PRESCALE_MASK (0)
#define ACKN_TIMER_PRESCALE      (1<<ACKN_TIMER_PRESCALE_MASK)
#define ACKN_TIMER_FREQ          (BUS_FREQ/ACKN_TIMER_PRESCALE)
#define ACKN_MICROSECOND(x)      ((U16)(((x)*(ACKN_TIMER_FREQ/1000))/1000UL))  // SYNC timer ticks in 1 us
#define ACKN_TPM1SC_VALUE        (TPM1SC_CLKSA_MASK|ACKN_TIMER_PRESCALE_MASK)

// TPM2 Fast timer constants ~333ns resolution, usable to 20ms
#define FAST_TIMER_PRESCALE_MASK (3)
#define FAST_TIMER_PRESCALE      (1<<FAST_TIMER_PRESCALE_MASK)
#define FAST_TIMER_FREQ          (BUS_FREQ/FAST_TIMER_PRESCALE)
#define MICROSECOND(x)           ((U16)(((x)*(FAST_TIMER_FREQ/1000))/1000UL))  // Fast timer ticks in 1 us
#define FAST_TPMxSC_VALUE        (TPM2SC_CLKSA_MASK|FAST_TIMER_PRESCALE_MASK)

// TPM2 Slow timer constants ~5us resolution, usable to 320ms
#define SLOW_TIMER_PRESCALE_MASK (FAST_TIMER_PRESCALE_MASK+4)
#define SLOW_TIMER_PRESCALE      (1<<SLOW_TIMER_PRESCALE_MASK)
#define SLOW_TIMER_FREQ          (BUS_FREQ/SLOW_TIMER_PRESCALE)
#define MILLISECOND(x)           ((U16)(((x)*SLOW_TIMER_FREQ)/1000UL))  // Slow timer ticks in 1 ms
#define SLOW_TPMxSC_VALUE        (TPM2SC_CLKSA_MASK|SLOW_TIMER_PRESCALE_MASK)

// TIM-Chz Timer expired
#define WAIT_TIMER_EXPIRED       (TPM2SC_TOF!=0)

#if (SLOW_TIMER_PRESCALE_MASK >= 8)
#error "Timer prescale value out of range"
#endif

//! Wait for given time in fast timer ticks
//! 
//!  @param delay Delay time in fast timer ticks
//!
static void fastTimerWait(U16 delay){
   TPM2SC      = 0;                             // Set time out value, Immediate change
   TPM2MOD     = delay;                         //  - requires CLKSA:CLKSB=00
   TPM2SC      = FAST_TPMxSC_VALUE;             // Set timer count rate
   TPM2CNTH    = 0;                             // Restart timer counter
   TPM2C0SC    = TPM2C0SC_MS0A_MASK;            // Output compare function, pin disabled
   TPM2SC_TOF  = 0;                             // Clear flag
   while (!WAIT_TIMER_EXPIRED) {
   }
}

//! Wait for given time in slow timer ticks
//! 
//!  @param delay Delay time in slow timer ticks
//!
static void slowTimerWait(U16 delay){
   TPM2SC      = 0;                             // Set time out value
   TPM2MOD     = delay;                         //  - requires CLKSA:CLKSB=00
   TPM2SC      = SLOW_TPMxSC_VALUE;             // Set timer count rate
   TPM2CNTH    = 0;                             // Restart timer counter
   TPM2C0SC    = TPM2C0SC_MS0A_MASK;            // Output compare function, pin disabled
   TPM2SC_TOF  = 0;                             // Clear flag
   while (!WAIT_TIMER_EXPIRED) {
   }
}

/*! \brief A Macro to wait for given time (makes use of timer).

    @param  t  Time to wait in \e microseconds.
*/
#define WAIT_US(t) fastTimerWait(MICROSECOND(t))
/*! \brief A Macro to wait for given time (makes use of timer).

    @param  t  Time to wait in \e milliseconds.
*/
#define WAIT_MS(t) slowTimerWait(MILLISECOND(t))
/*! \brief A Macro to wait for given time (makes use of timer).

    @param  t  Time to wait in \e seconds.
*/
#define WAIT_S(t)  {int tt = 4*(t); \
                    while(tt-->0)   \
                       slowTimerWait(MILLISECOND(250));}

/*! \brief A Macro to wait for given time or until a condition is met

    @param  t  Maximum time to wait in \e microseconds.
    @param  c  Condition to exit wait early
*/
#define WAIT_WITH_TIMEOUT_US(t,c) {             \
   TPM2SC      = 0;                             \
   TPM2MOD     = MICROSECOND(t);                \
   TPM2SC      = FAST_TPMxSC_VALUE;             \
   TPM2CNTH    = 0;                             \
   TPM2C0SC    = TPM2C0SC_MS0A_MASK;            \
   TPM2SC_TOF  = 0;                             \
   while (!(c) && !WAIT_TIMER_EXPIRED) {        \
   }                                            \
}

/*! \brief A Macro to wait for given time or until a condition is met

    @param  t  Maximum time to wait in \e milliseconds.
    @param  c  Condition to exit wait early
*/
#define WAIT_WITH_TIMEOUT_MS(t,c) {             \
   TPM2SC      = 0;                             \
   TPM2MOD     = MILLISECOND(t);                \
   TPM2SC      = SLOW_TPMxSC_VALUE;             \
   TPM2CNTH    = 0;                             \
   TPM2C0SC    = TPM2C0SC_MS0A_MASK;            \
   TPM2SC_TOF  = 0;                             \
   while (!(c) && !WAIT_TIMER_EXPIRED) {        \
   }                                            \
}

/*! \brief A Macro to wait for given time or until a condition is met

    @param  t  Maximum time to wait in \e seconds.
    @param  c  Condition to exit wait early (checked every ~10 ms and affects timing)
*/
#define WAIT_WITH_TIMEOUT_S(t,c) {      \
    int tt = 100*(t);                   \
      do {                              \
         slowTimerWait(MILLISECOND(10)); \
      } while (!(c) & (tt-->0));        \
   }

#pragma MESSAGE DISABLE C5909 // Disable warnings about Assignment in condition
#pragma MESSAGE DISABLE C4000 // Disable warnings about Condition always true

//! Read Target BDM status
//!
//!  Depending on target architecture this reads from
//!  - BDCSC, 
//!  - BDMSTS,
//!  - XCSR.
//!
//! @param bdm_sts location to write status value
//! @return
//!   \ref BDM_RC_OK            => success  \n
//!   \ref BDM_RC_UNKNOWN_TARGET => unknown target
//!
U8 bdm_readBDMStatus(U8 *bdm_sts) {

   switch (bdm_status.target_type) {
      case HC12:
         BDM12_CMD_BDREADB(HC12_BDMSTS,bdm_sts);
         return BDM_RC_OK;
      case HCS08:
      case RS08:
         BDM08_CMD_READSTATUS(bdm_sts);
         return BDM_RC_OK;
      case CFV1:
         BDMCF_CMD_READ_XCSR(bdm_sts);
         return BDM_RC_OK;
      default:
         return BDM_RC_UNKNOWN_TARGET; // Don't know how to check status on this one!
   }
}

//! Write Target BDM control register 
//!
//!  Depending on target architecture this writes to \n
//!   - BDCSC,
//!   - BDMSTS,
//!   - XCSR.
//!
//! @return
//!   \ref BDM_RC_OK            => success  \n
//!   \ref BDM_RC_UNKNOWN_TARGET => unknown target
//!
U8 bdm_writeBDMControl(U8 bdm_ctrl) {
U8 statusClearMask; // Bits to clear in control value
U8 statusSetMask;   // Bits to set in control value
U8 statusClkMask;   // The position of the CLKSW bit in control register

   // Get clock select mask for this target (CLKSW bit in BDM control register)
   switch (bdm_status.target_type) {
      case HC12:
         statusClkMask = HC12_BDMSTS_CLKSW;
         break;
      case HCS08:
      case RS08:
         statusClkMask = HC08_BDCSCR_CLKSW;
         break;
      case CFV1:
         statusClkMask = CFV1_XCSR_CLKSW;
         break;
      default:
         return BDM_RC_UNKNOWN_TARGET; // Don't know how to check status on this one!
   }

   // Construct the masks to modify control value
   switch (bdm_option.useAltBDMClock) {
      case ALT:     statusClearMask = ~statusClkMask;  statusSetMask   = 0x00;          break; // CLKSW = 0
      case NORMAL:  statusClearMask = 0xFF;            statusSetMask   = statusClkMask; break; // CLKSW = 1
      default:      statusClearMask = 0xFF;            statusSetMask   = 0x00;          break; // CLKSW whatever given
   }
   
   switch (bdm_status.target_type) {
      case HC12:
         BDM12_CMD_BDWRITEB(HC12_BDMSTS,(bdm_ctrl&statusClearMask)|statusSetMask);
         return BDM_RC_OK;
      case HCS08:
      case RS08:
         BDM08_CMD_WRITECONTROL((bdm_ctrl&statusClearMask)|statusSetMask);
         return BDM_RC_OK;
      case CFV1:
         BDMCF_CMD_WRITE_XCSR((bdm_ctrl&statusClearMask)|statusSetMask);
         return BDM_RC_OK;
   }
}

//! If BDM mode is not enabled in target yet, enable it so it can be made active
//!
//! @return
//!   \ref BDM_RC_OK            => success  \n
//!    \ref BDM_RC_UNKNOWN_TARGET => unknown target \n
//!    \ref BDM_RC_BDM_EN_FAILED  => enabling BDM failed (target not connected or wrong speed ?)
//!
U8 bdm_enableBDM() {
U8 bdm_sts;
U8 enbdmMask;
U8 rc;

   if (bdm_status.target_type==CFV1) // Choose appropriate mask
      enbdmMask = CFV1_XCSR_ENBDM;
   else
      enbdmMask = HC12_BDMSTS_ENBDM;

   rc = bdm_readBDMStatus(&bdm_sts); // Get current status
   if (rc != BDM_RC_OK)
      return rc;
      
   if ((bdm_sts & enbdmMask) != 0) // BDM already enabled?
      return BDM_RC_OK;
      
   rc = bdm_writeBDMControl(bdm_sts | enbdmMask); // Try to enable
   if (rc != BDM_RC_OK)
      return rc;
      
   rc = bdm_readBDMStatus(&bdm_sts); // Enabled now?
   if (rc != BDM_RC_OK)
      return rc;
      
   return ((bdm_sts & enbdmMask)!=0)? 0 : BDM_RC_BDM_EN_FAILED;

   }
#pragma MESSAGE DEFAULT C4000 // Restore warnings about Condition always true
#pragma MESSAGE DEFAULT C5909 // Restore warnings about Assignment in condition

//! Used before some commands to ensure there is a useful connection to target available.
//!
//! Some non-intrusive commands still require a running processor (e.g. READ_BYTE)
//! This function will force a stopped CPU into active BDM mode 
//!
//! @return
//!   \ref BDM_RC_OK            => success  \n
//!   otherwise  => various errors
//!
U8 bdm_makeActiveIfStopped(void) {

U8 rc;
U8 bdm_sts;

   if (bdm_status.target_type == HC12) // Not supported on HC12
      return BDM_RC_OK; // not considered an error

   rc = 0;
   if (bdm_option.autoReconnect &&           // Auto re-connect enabled &
       (bdm_status.speed==SYNC_SUPPORTED)) { // Target supports Sync
      rc = bdm_syncMeasure(); // Check connection speed
      if (rc == 0)
         rc = bdm_RxTxSelect();
      }
   if (rc != BDM_RC_OK)
      return rc;

   rc = bdm_readBDMStatus(&bdm_sts);
   if (rc != BDM_RC_OK)
      return rc;

   // If HC08/RS08 processor is WAITing or STOPped then force background
   if (bdm_sts&HC08_BDCSCR_WS)
      bdm_halt();

   return BDM_RC_OK;
}

//!  Tries to connect to target - doesn't try other strategies such as reset
//!
//!  This function does a basic connect sequence.  It doesn't configure the BDM registers on the target
//!
//! @return
//!    \ref    BDM_RC_OK               => success  \n
//!    \ref    BDM_RC_VDD_NOT_PRESENT  => no target power present \n
//!    \ref    BDM_RC_RESET_TIMEOUT    => RESET signal timeout (remained low)
//!    \ref    BDM_RC_BKGD_TIMEOUT     => BKGD signal timeout (remained low)
//!    != \ref BDM_RC_OK               => other failures
//!
static U8 bdm_physicalConnect(void){
       U8 rc;

   bdm_status.ackn  = WAIT;              // Clear the ACKN feature
//   bdm_status.reset = NO_RESET_ACTIVITY; // Clear the reset flag
   bdm_status.ackn  = NO_INFO;           // No connection
   bdm_rx_ptr       = bdm_rxEmpty;       // Clear the Tx/Rx pointers
   bdm_tx_ptr       = bdm_txEmpty;       //    i.e. no com. routines found

   bdm_idle(); // Make sure interface is idle

   // Target has power?
   if (!bdm_status.power)
      return BDM_RC_VDD_NOT_PRESENT;
      
   // Wait with timeout until both RESET and BKGD  are high
#if (CAPABILITY&CAP_RESET)
   if (bdm_option.useResetSignal) {
      WAIT_WITH_TIMEOUT_MS(RESET_WAIT,(RESET_IN!=0)&&(BDM_IN!=0));
      if (RESET_IN==0) 
         return(BDM_RC_RESET_TIMEOUT);  // RESET timeout
   }
#else   
   WAIT_WITH_TIMEOUT_MS(RESET_WAIT,(BDM_IN!=0));
#endif   
      
   if (BDM_IN==0) 
      return(BDM_RC_BKGD_TIMEOUT);  // BKGD timeout

   rc = bdm_syncMeasure();
   if ((rc != BDM_RC_OK) &&                // Trying to measure SYNC was not successful
       (bdm_option.guessSpeed) &&          // Try alternative method if enabled
       (bdm_status.target_type == HC12)) { // and HC12 target
      rc = bdmHC12_alt_speed_detect();     // Try alternative method (guessing!)
   }
   if (rc != BDM_RC_OK)
      return(rc);

   // If at least one of the two methods succeeded, we can select
   //  the right Rx and Tx routines
   rc = bdm_RxTxSelect();
   if (rc != BDM_RC_OK)
      return(rc);
      
   bdm_acknInit();  // Try the ACKN feature
   
   return BDM_RC_OK;
}

//! Connect to target
//!
//!  This function may cycle the target power in attempting to connect. \n
//!  It enable BDM on the target if connection is successful.
//!
//! @return
//!    \ref    BDM_RC_OK               => success  \n
//!    \ref    BDM_RC_VDD_NOT_PRESENT  => no target power present \n
//!    \ref    BDM_RC_RESET_TIMEOUT    => RESET signal timeout (remained low)
//!    \ref    BDM_RC_BKGD_TIMEOUT     => BKGD signal timeout (remained low)
//!    != \ref BDM_RC_OK               => other failures
//!
U8 bdm_connect(void) {
U8 rc;
U8 bdm_sts;

   rc = bdm_physicalConnect();
   
   if ((rc != BDM_RC_OK) && bdm_option.cycleVddOnConnect) { // No connection to target - cycle power if allowed
      (void)bdm_cycleTargetVdd(); // Ignore errors
      rc = bdm_physicalConnect(); // Try connect again
   }
   
   if (rc != BDM_RC_OK)
      return rc;

   /* try to enable BDM */
   rc = bdm_enableBDM();
   if (rc != BDM_RC_OK)
      return(rc);

   // If HC08/RS08 processor is halted then force background
   rc = bdm_readBDMStatus(&bdm_sts);
   if (rc != BDM_RC_OK)
      return rc;

   rc = bdm_writeBDMControl(bdm_sts); // In case Force BDM clock is active
   
//      
//   if ((bdm_status.target_type != HC12) && (bdm_sts&HC08_BDCSCR_WS))
//      bdm_halt();

   return rc;
}

//!  Mass erase (RS08)
//!
//! @entry
//!     VPP is already on
//!
//! @return
//!   \ref BDM_RC_OK            => success
//!
U8  bdmRS08_massErase(void) {
U16 FLCRaddr;
U8 rc;

   rc = bdm_connect();
   if (rc != BDM_RC_OK)
      return rc;
   
   bdm_halt();

   switch (bdm_option.derivative_type) {
      case KA1:
      case KA2:
         FLCRaddr = RS08KA2_FLCR;
         break;
      case KA4:
      case KA8:
         FLCRaddr = RS08KA8_FLCR;
         break;
      case LA8:
         FLCRaddr = RS08LA8_FLCR;
         break;
      default:
         return BDM_RC_UNKNOWN_TARGET;
   }

   // DisableInterrupts;

   // Retrieve and save the trim value from RS08 Flash
   BDM08_CMD_READB(RS_TRIM_FLASH,&savedTrimValue);

   // Set Up PAGESEL to point to end of flash
   BDM08_CMD_WRITEB(RS_PAGESEL,0xfe);

   // Set the Flash Mass Erase Bit
   BDM08_CMD_WRITEB(FLCRaddr,RS_MASS);

   // Access the Flash through the paging window	
   BDM08_CMD_WRITEB(RS_PAGEWIN+1,0x55);

   // Set the HVEN Bit
   BDM08_CMD_WRITEB(FLCRaddr,RS_MASS|RS_HVEN);

   // Wait 500ms
   WAIT_MS(250);
   WAIT_MS(250);

   // Clear the Mass Bit
   BDM08_CMD_WRITEB(FLCRaddr,RS_HVEN);

   WAIT_MS(100);

   // Clear HVEN
   BDM08_CMD_WRITEB(FLCRaddr,0x00);

   (void)bdm_softwareReset(0);

   bdm_halt();
   
   return BDM_RC_OK;
}

//=======================================================================
//  Memory Map for RS08 Flash programming variables
//
//      ;********************************************************************************
//      ; Fast RAM ($0 - $D)
//                        org   $0
//      ;
//      ; Programming variables - re-initialise each call
//      ;
// 0000 SourceAddr        ds.b  1 ; Address/Count in Zero Page (Buffer) for source
// 0001 Buffer            ds.b  8 ; Buffer for data to program (reverse order)
// 0009 DestinationAddr   ds.b  1 ; Address in Zero Page (Paging Window) for destination
// 000A DestinationPage   ds.b  1 ; Page for destination
//      ;
//      ; Timing parameters for Flash Programming - constants
//      ;
// 000B TimeTnvs          ds.b  1
// 000C TimeTpgs          ds.b  1
// 000D TimeTnvh          ds.b  1
//

//!  Programs a block of up to 8 bytes in the RS08 Flash.
//!
//! @entry
//!  It assumes that the Flash programming code & data have been downloaded
//!  into the RS08 RAM previously . \n
//!  It then tells the RS08 to execute the program in RAM and waits
//!  for completion. \n
//!  Assumes VPP is already on
//!
static void bdmRS08_programFlashBlock(void) {
U8 bdm_sts;
U8 count;

   // Set the PC to start of Flash programming code in RAM     
   switch (bdm_option.derivative_type) {
      case KA1:
      case KA2:
         BDM08_CMD_WRITE_PC(RS08KA2_DIR_RAM);
         break;
      case KA4:
      case KA8:
         BDM08_CMD_WRITE_PC(RS08KA8_DIR_RAM);
         break;
      default:
         //rc = BDM_RC_UNKNOWN_TARGET;
         return;
   }
   // We are now ready to program - execute the code on RS08
   bdm_go();

   // Wait until RS08 completes execution and re-enters BGND mode
   // Timeout accounts for a maximum of 8 bytes of programming, 400us ~ 40us * 8 + 33%
//   WAIT_WITH_TIMEOUT_US(400,((bdm_readBDMStatus(&bdm_sts) == 0) && ((bdm_sts&HC08_BDCSCR_BDMACT)!=0)));
   for (count = 8; count-- > 0;) {
      //enableInterrupts();
      WAIT_US(55 /* us*/);
      disableInterrupts();
      if ((bdm_readBDMStatus(&bdm_sts) == 0) && ((bdm_sts&HC08_BDCSCR_BDMACT)!=0))
         break;
   }
}

//!  Program Target Clock Osc Trim value in Flash
//!
//!  @entry
//!     Flash programming code is already in RS08 RAM \n
//!     Flash has already been programmed with user code \n
//!     VPP is on \n
//!
//!  Will used saved Clock Trim value or calculate new Trim as necessary
//!
void bdmRS08_restoreTrim(void) {
   U8 currentTrimValue;

   // Check if Trim location already programmed as part of downloaded image
   BDM08_CMD_READB(RS_TRIM_FLASH,&currentTrimValue);
   if (currentTrimValue != 0xff)
      return;

   // If there is no saved trim value, or a trim frequency has been supplied then calculate new value trim
   if ((savedTrimValue == 0xFF) || (bdm_option.targetClockTrim != 0)) {
      if (bdm_option.targetClockTrim != 0) // Clock trim SYNC value has been supplied
         (void)bdmRS08_trimTargetClock(bdm_option.targetClockTrim);
      else // Use default trim frequency of 8MHz
         (void)bdmRS08_trimTargetClock(SYNC_MULTIPLE(8000000UL));
   }
   
   currentTrimValue = savedTrimValue; // Save calculated trim value
   
   // Bit messy - we have to trim the clock for programming again in case the above upset it
   (void)bdmRS08_trimTargetClock(SYNC_MULTIPLE(8000000UL));
   
   // # of bytes to program
   BDM08_CMD_WRITEB(RS_FAST_RAM+0, 1);

   // Fill buffer (1 byte!)
   BDM08_CMD_WRITEB(RS_FAST_RAM+1, currentTrimValue);

   // Address in Zero Page (Paging Window) for destination
   BDM08_CMD_WRITEB(RS_FAST_RAM+0x09, RS08_WIN_ADDR(RS_TRIM_FLASH));

   // Page for destination
   BDM08_CMD_WRITEB(RS_FAST_RAM+0x0A, RS08_PAGENO(RS_TRIM_FLASH));

   bdmRS08_programFlashBlock();
}

//!  Download program/data into the RS08 RAM
//!
//! @entry
//!  Assumes target connection
//!
static void bdmRS08_loadRAMProgram(const U8 *programCode){
   U16 addr;
   U8  count;
   
   // Enter Active BG mode change speed and re/sync
   bdm_halt();

   count = *programCode++;                         // Size of program to load
   addr  = *(U16*)programCode;  programCode+=2;    // RAM address to load program
   
   // Download Flash Programming code to RS08 Direct RAM
   while (count-->0) {
      BDM08_CMD_WRITEB(addr,*programCode);
      addr++;
      programCode++ ;
      }
}

#pragma MESSAGE DISABLE C5900 // Disable warnings about Result is zero

//!  This routine programs a block of up to 64 bytes into the RS08
//!
//!  Method:
//!  * Initially a program is downloaded into the RS08 RAM.  This program
//!    can program up to 8-bytes of data into the Flash at a time.
//!  * The data is then downloaded in small blocks and the program used
//!    to program it into the Flash memory.
//!  This process is repeated until the whole block is programmed into the Flash
//!
//! @entry
//!  Assumes VPP is already on \n
//!  Uses \ref flashProgrammingBuffer[] as follows:
//!     - [3]:[2] = Start address (little-endian)
//!     - [4]     = # of bytes to program (<= 64 bytes)
//!     - [5...]  = data to program
//!
U8 bdmRS08_programFlash(void) {
   U8  dataPtr;
   U16 addr;
   U8  bytesRemaining;
   U8  bytesThisBlock;
   U8  rc;
 
   // Trim target clock to 8MHz as assumed for programming
   rc = bdmRS08_trimTargetClock(SYNC_MULTIPLE(8000000UL));
   if (rc != BDM_RC_OK)
      return rc;
   
   bdm_halt();

   // Download appropriate Flash Programming code to RS08 Direct RAM
   switch (bdm_option.derivative_type) {
      case KA1:
      case KA2:
         bdmRS08_loadRAMProgram(RS08KA2_Flash_Program);
         break;
      case KA4:
      case KA8:
         bdmRS08_loadRAMProgram(RS08KA8_Flash_Program);
         break;
      case LA8:
         bdmRS08_loadRAMProgram(RS08LA8_Flash_Program);
         break;
      default:
         return BDM_RC_UNKNOWN_TARGET;
   }
   
   // Enter Active BG mode
   bdm_halt();

   // Set timing Constants
   BDM08_CMD_WRITEB(RS_FAST_RAM+0x0B,  4); // Tnvs, Tnvh = 5 us
   BDM08_CMD_WRITEB(RS_FAST_RAM+0x0C,  6); // Tpgs = 10 us
   BDM08_CMD_WRITEB(RS_FAST_RAM+0x0D, 23); // Tprog = 40 us

   // Address to start programming at (littleEndian)
   addr    = (flashProgrammingBuffer[3]<<8)+flashProgrammingBuffer[2];

   // Offset in buffer of data to program
   dataPtr = 5;

   // Number of bytes remaining to program
   bytesRemaining = flashProgrammingBuffer[4];

   while ( bytesRemaining > 0) { 

      // Program in blocks of up to 8 bytes within an 8-byte boundary
      bytesThisBlock = 8 - (addr & 7);

      // Check if this is more than remaining bytes to program
      if (bytesRemaining < bytesThisBlock)
         bytesThisBlock = bytesRemaining;

      // Set # of bytes to program
      BDM08_CMD_WRITEB(RS_FAST_RAM+0x00,bytesThisBlock);

      // Address in Zero Page (within Paging Window) for destination
      BDM08_CMD_WRITEB(RS_FAST_RAM+0x09,RS08_WIN_ADDR(addr));

      // Page for destination
      BDM08_CMD_WRITEB(RS_FAST_RAM+0x0A,RS08_PAGENO(addr));

      bytesRemaining -= bytesThisBlock;
      addr           += bytesThisBlock;

      do {
         // Copy data to buffer in RS08 (in reverse order!)
         BDM08_CMD_WRITEB(RS_FAST_RAM+bytesThisBlock,flashProgrammingBuffer[dataPtr]);
         dataPtr++;
         } while (--bytesThisBlock > 0);

      bdmRS08_programFlashBlock(); // Program up to 8 bytes
   }
   
   (void)bdm_softwareReset(0);

   bdm_halt();
   //bdm_connect();
   
   return BDM_RC_OK;
}
#pragma MESSAGE DEFAULT C5900 // Restore warnings about Result is zero

#define VDD_2v  (((2*255)/5)*9/10)  // 10% allowance on 2V
#define VDD_3v3 (((3*255)/5)*9/10)  // 10% allowance on 3.3V
#define VDD_5v  (((5*255)/5)*9/10)  // 10% allowance on 5V

//! Turns on Target Vdd if enabled.
//!
//!  @return
//!   \ref BDM_RC_OK  => Target Vdd confirmed on target \n
//!   \ref BDM_RC_VDD_NOT_PRESENT   => Target Vdd not present
//!
static U8 bdm_setTargetVdd( void ) {
U8 rc = 0;
#if (CAPABILITY&CAP_VDDCONTROL)
    switch (bdm_option.targetVdd) {
      case BDM_TARGET_VDD_OFF :
         VDD_OFF();
         // Check for externally supplied target Vdd (> 2 V)
         WAIT_US(VDD_RISE_TIME); // Wait for Vdd to rise & stabilise
         if (bdm_targetVddMeasure()<VDD_2v)
            rc = BDM_RC_VDD_NOT_PRESENT;
         break;
      case BDM_TARGET_VDD_3V3 :
         VDD3_ON();
         // Wait for Vdd to rise to 90% of 3V
         WAIT_WITH_TIMEOUT_MS( 250 /* ms */, (bdm_targetVddMeasure()>VDD_3v3));
         WAIT_US(VDD_RISE_TIME); // Wait for Vdd to rise & stabilise
         if (bdm_targetVddMeasure()<VDD_3v3) {
            VDD_OFF(); // In case of Vdd overload
            rc = BDM_RC_VDD_NOT_PRESENT;
            }
         break;
      case BDM_TARGET_VDD_5V  :
         VDD5_ON();
         // Wait for Vdd to rise to 90% of 5V
         WAIT_WITH_TIMEOUT_MS( 250 /* ms */, (bdm_targetVddMeasure()>VDD_5v));
         WAIT_US(VDD_RISE_TIME); // Wait for Vdd to rise & stabilise
         if (bdm_targetVddMeasure()<VDD_5v) {
            VDD_OFF(); // In case of Vdd overload
            rc = BDM_RC_VDD_NOT_PRESENT;
            }
         break;
    }
#endif // CAP_VDDCONTROL

#if (CAPABILITY&CAP_VDDSENSE)
   CLEAR_VDD_SENSE_FLAG(); // Clear Vdd Change Event
#endif

   bdm_checkTargetVdd(); // Update Target Vdd LED
   
   return (rc);
}

//!  Cycle power ON to target
//!
//!  BKGD is held low when power is re-applied to start
//!  target with BDM active
//!
//!  @return
//!   \ref BDM_RC_OK  => Target Vdd confirmed on target \n
//!   \ref BDM_RC_VDD_WRONG_MODE    => Target Vdd not controlled by BDM interface \n
//!   \ref BDM_RC_VDD_NOT_PRESENT   => Target Vdd failed to rise \n
//!   \ref BDM_RC_RESET_TIMEOUT     => RESET signal failed to rise \n
//!   \ref BDM_RC_BKGD_TIMEOUT      => BKGD signal failed to rise
//!
U8 bdm_cycleTargetVddOn(void) {
U8 rc = 0;

#if (CAPABILITY&CAP_VDDCONTROL)

   bdm_idle();  // Make sure BDM interface is idle
      
#if (DEBUG&CYCLE_DEBUG)
   DEBUG_PIN     = 0;
   DEBUG_PIN     = 1;
#endif

   // BKGD pin=L
   DATA_PORT      = BDM_DIR_Rx_WR;
   DATA_PORT_DDR  = BDM_DIR_Rx_MASK|BDM_OUT_MASK;
//   RESET_LOW();

#if (DEBUG&CYCLE_DEBUG)
   DEBUG_PIN     = 0;
   DEBUG_PIN     = 1;
#endif

   // Power on (BKGD Low)
   rc = bdm_setTargetVdd();
   if (rc != BDM_RC_OK) // No target Vdd
      goto cleanUp;

#if (DEBUG&CYCLE_DEBUG)
   DEBUG_PIN     = 1;
   DEBUG_PIN     = 0;
#endif

//   RESET_3STATE();
   // RESET rise may be delayed by target POR
   if (bdm_option.useResetSignal)
      WAIT_WITH_TIMEOUT_MS( 250 /* ms */, (RESET_IN!=0) );

#if (DEBUG&CYCLE_DEBUG)
   DEBUG_PIN   = 0;
   DEBUG_PIN   = 1;
#endif // (DEBUG&CYCLE_DEBUG)

   // Let signals settle & CPU to finish reset (with BKGD held low)
   WAIT_US(BKGD_WAIT);

#if (CAPABILITY&CAP_RESET)
   if (bdm_option.useResetSignal && (RESET_IN==0)) { 
      // RESET didn't rise
      rc = BDM_RC_RESET_TIMEOUT;
      goto cleanUp;
      }
#endif //(CAPABILITY&CAP_RESET)

#if (DEBUG&CYCLE_DEBUG)
   DEBUG_PIN     = 1;
   DEBUG_PIN     = 0;
#endif // (DEBUG&CYCLE_DEBUG)

   bdm_idle();  // Release BKGD

   // Let processor start up
   WAIT_MS(RESET_RECOVERY);

   if (BDM_IN==0) { // BKGD didn't rise!
      rc = BDM_RC_BKGD_TIMEOUT;
      goto cleanUp;
      }

   bdm_status.reset  = RESET_DETECTED; // Cycling the power should have reset it!
      
cleanUp:
   bdm_idle();  // Make sure BDM interface is idle
   WAIT_MS( 250 /* ms */);

#endif // CAP_VDDCONTROL

   return(rc);
}

//!  Cycle power OFF to target
//!
//!  @return
//!   \ref BDM_RC_OK  => No error  \n
//!   \ref BDM_RC_VDD_WRONG_MODE    => Target Vdd not controlled by BDM interface \n
//!   \ref BDM_RC_VDD_NOT_REMOVED   => Target Vdd failed to fall \n
//!
U8 bdm_cycleTargetVddOff(void) {
U8 rc = 0;

#if (CAPABILITY&CAP_VDDCONTROL)

   bdm_checkTargetVdd();

   if (bdm_option.targetVdd == BDM_TARGET_VDD_OFF)
      return BDM_RC_VDD_WRONG_MODE;

   bdm_idle();  // Make sure BDM interface is idle
      
#if (DEBUG&CYCLE_DEBUG)
   DEBUG_PIN     = 0;
   DEBUG_PIN     = 1;
#endif

   // BKGD pin=L, RESET pin=L
   DATA_PORT      = BDM_DIR_Rx_WR;
   DATA_PORT_DDR  = BDM_DIR_Rx_MASK|BDM_OUT_MASK;
   RESET_LOW();

   // Power off & wait for Vdd to fall to ~5%
   VDD_OFF();
   WAIT_WITH_TIMEOUT_S( 5 /* s */, (bdm_targetVddMeasure()<10) );

#if (DEBUG&CYCLE_DEBUG)
   DEBUG_PIN   = 1;
   DEBUG_PIN   = 0;
#endif
   
   if (bdm_targetVddMeasure()>=15) // Vdd didn't turn off!
      rc = BDM_RC_VDD_NOT_REMOVED;

#if (DEBUG&CYCLE_DEBUG)
   DEBUG_PIN     = 0;
   DEBUG_PIN     = 1;
#endif

   bdm_checkTargetVdd(); // Update Target Vdd LED

   // Wait a while with power off
   WAIT_US(RESET_SETTLE);
      
   bdm_idle();  // Make BDM interface idle

   // Clear Vdd monitoring interrupt
#if (CAPABILITY&CAP_VDDSENSE)
   CLEAR_VDD_SENSE_FLAG();  // Clear Vdd monitoring flag
#endif   
   bdm_checkTargetVdd(); // Update Target Vdd LED

#endif // CAP_VDDCONTROL

   return(rc);
}

//!  Cycle power to target
//!
//!  BKGD is held low when power is re-applied to start
//!  target with BDM active
//!
//!  @return
//!   \ref BDM_RC_OK                => No error \n
//!   \ref BDM_RC_VDD_WRONG_MODE    => Target Vdd not controlled by BDM interface \n
//!   \ref BDM_RC_VDD_NOT_REMOVED   => Target Vdd failed to fall \n
//!   \ref BDM_RC_VDD_NOT_PRESENT   => Target Vdd failed to rise \n
//!   \ref BDM_RC_RESET_TIMEOUT     => RESET signal failed to rise \n
//!   \ref BDM_RC_BKGD_TIMEOUT      => BKGD signal failed to rise
//!
U8 bdm_cycleTargetVdd(void) {
U8 rc = 0;

   rc = bdm_cycleTargetVddOff();
   if (rc != BDM_RC_OK)
      return(rc);
   rc = bdm_cycleTargetVddOn();
   return(rc);
}

#if (CAPABILITY&CAP_RESET)

//! Resets the target using the BDM reset line.
//! @note
//!   Not all targets have a reset pin or support reset to BDM mode with a reset pin.
//!
//! @param mode
//!    - SPECIAL_MODE => Reset to special mode,
//!    - NORMAL_MODE  => Reset to normal mode
//!
//! @return
//!   \ref BDM_RC_OK  => Success \n
//!   \ref BDM_RC_BKGD_TIMEOUT     => BKGD pin stuck low \n
//!   \ref BDM_RC_RESET_TIMEOUT    => RESET pin stuck low \n
//!
U8 bdm_hardwareReset(U8 mode) {

   DISABLE_RESET_SENSE_INT(); // Mask RESET IC interrupts
   
   bdm_idle();  // Make sure BDM interface is idle
   
   if (!bdm_option.useResetSignal)
      return BDM_RC_ILLEGAL_COMMAND;

   // Wait with timeout until both RESET and BKGD are high
   WAIT_WITH_TIMEOUT_MS(RESET_WAIT, (RESET_IN!=0)&&(BDM_IN!=0));

   if (RESET_IN==0) return(BDM_RC_RESET_TIMEOUT);
   if (BDM_IN==0)   return(BDM_RC_BKGD_TIMEOUT);

   bdm_txPrepare(); // BKGD & RESET 3-state control on DIR_PORT

   if (mode==SPECIAL_MODE) {
      BDM_OUT = 0;        // Drive BKGD low 
      BDM_ENABLE_TX();     
      
      WAIT_US(RESET_SETTLE); // Wait for signals to settle
   }

   RESET_LOW();

   WAIT_MS(RESET_LENGTH);  // Wait for reset pulse duration

   RESET_3STATE();

   // Wait with timeout until RESET is high
   WAIT_WITH_TIMEOUT_MS(RESET_WAIT, (RESET_IN!=0));

   // Assume RESET risen - check later after cleanUp

#if (DEBUG&RESET_DEBUG)
   DEBUG_PIN   = 0;
   DEBUG_PIN   = 1;
#endif

   // wait recovery time before allowing anything else to happen on the BDM
   WAIT_MS(RESET_RECOVERY);  // Wait for Target to start up after reset

   bdm_idle(); // Place interface in idle state

#if (DEBUG&RESET_DEBUG)
   DEBUG_PIN   = 1;
   DEBUG_PIN   = 0;
#endif

   if (RESET_IN==0)      // RESET failed to rise
      return(BDM_RC_RESET_TIMEOUT);
   
   CLEAR_RESET_SENSE_FLAG(); // Clear RESET IC Event
   ENABLE_RESET_SENSE_INT(); // Enable RESET IC interrupts

   return(BDM_RC_OK);
}

#endif //(CAPABILITY&CAP_RESET)
   
//! Resets the target using BDM commands
//!
//! @note
//!    Not all targets support reset using BDM commands
//!
//! @param mode
//!    - SPECIAL_MODE => Reset to special mode,
//!    - NORMAL_MODE => Reset to normal mode
//!
//! @return
//!   \ref BDM_RC_OK  => Success \n
//!    \ref BDM_RC_BKGD_TIMEOUT     => BKGD pin stuck low \n
//!    \ref BDM_RC_RESET_TIMEOUT    => RESET pin stuck low \n
//!    \ref BDM_RC_UNKNOWN_TARGET   => Don't know how to reset this type of target! \n
//!
U8 bdm_softwareReset(U8 mode) {
U8 rc;

   if (bdm_status.target_type == HC12) // Doesn't support s/w reset
      return BDM_RC_UNKNOWN_TARGET;

   // Make sure of connection
   rc = bdm_connect();

   // Make sure Active background mode (in case target is stopped!)
   bdm_halt();

   switch (bdm_status.target_type) {
      case HCS08:
         BDM08_CMD_RESET();
         break;
      case RS08:
         BDMRS08_CMD_RESET();
         break;
      case CFV1:
         // Force reset (& set up to halt in BDM mode on various errors)
         if (mode == 0) {
            BDM_CMD_1B_0_T(_BDMCF_WRITE_CSR2_BYTE, 
                  CFV1_CSR2_BDFR|CFV1_CSR2_COPHR|CFV1_CSR2_IOPHR|CFV1_CSR2_IADHR|CFV1_CSR2_BFHBR);
         }
         else {
            BDM_CMD_1B_0_T(_BDMCF_WRITE_CSR2_BYTE,
                  CFV1_CSR2_BDFR|CFV1_CSR2_COPHR|CFV1_CSR2_IOPHR|CFV1_CSR2_IADHR);
         }
         break;
      default:
         return BDM_RC_UNKNOWN_TARGET; // Don't know how to reset this one!
   }

   // BDM Interface is left in Tx mode by the above reset command

   if (mode == SPECIAL_MODE) {      // Special mode - need BKGD held low out of reset
      BDM_OUT  = 0;      // drive BKGD low (out of reset)
      BDM_ENABLE_TX();
   }
   
#if (DEBUG&RESET_DEBUG)
   DEBUG_PIN     = 0;
   DEBUG_PIN     = 1;
#endif

   WAIT_US(RESET_SETTLE);   // Wait for target to start reset (and possibly assert reset)

#if (CAPABILITY&CAP_RESET)
   if (bdm_option.useResetSignal) {
      // Wait with timeout until RESET is high (may be held low by processor)
      WAIT_WITH_TIMEOUT_MS(RESET_WAIT, (RESET_IN!=0));
      // Assume RESET risen - check later after cleanup
      }
#endif // (CAPABILITY&CAP_RESET)

#if (DEBUG&RESET_DEBUG)
   DEBUG_PIN   = 1;
   DEBUG_PIN   = 0;
#endif

   if (mode == 0) { // Special mode - release BKGD
      WAIT_US(BKGD_WAIT);        // Wait for BKGD assertion time after reset rise
      BDM_3STATE_TX();
      WAIT_US(RESET_SETTLE);     // time to wait for signals to settle
   }

   bdm_txFinish(); 

#if (DEBUG&RESET_DEBUG)
   DEBUG_PIN   = 0;
   DEBUG_PIN   = 1;
#endif
   
   // wait recovery time before allowing anything else to happen on the BDM
   WAIT_MS(RESET_RECOVERY);               // Wait for Target to start up after reset

   bdm_idle();                            // Place interface in idle state

#if (DEBUG&RESET_DEBUG)
   DEBUG_PIN   = 1;
   DEBUG_PIN   = 0;
#endif

   //  bdm_halt();  // For RS08?
   return BDM_RC_OK;
}

//! Resets the target
//!
//! Resets the target using any of the following as needed:
//! - Vdd power cycle (POR),
//! - Software BDM command or
//! - Hardware reset.
//!
//! @param mode
//!    - SPECIAL_MODE => Reset to special mode,
//!    - NORMAL_MODE => Reset to normal mode
//!
//! @return
//!     \ref BDM_RC_OK  => Success \n
//!    != \ref BDM_RC_OK     => various errors
//      
U8 bdm_targetReset( U8 mode ) {
U8 rc = BDM_RC_OK;

   // Power-on-reset - applies to all chips
   if (bdm_option.cycleVddOnReset)
      rc = bdm_cycleTargetVdd();

   if (rc != BDM_RC_OK)
      return rc;
     
   // Software (BDM Command) reset - HCS08, RS08 & Coldfire
   rc = bdm_softwareReset(mode);

#if (CAPABILITY&CAP_RESET)
   // Hardware (RESET pin) reset
   // HC12s and some HCS08s/RS08s & CFv1 (may not result in BDM mode) support this
   if ((rc != BDM_RC_OK) && bdm_option.useResetSignal)
      rc = bdm_hardwareReset(mode); 
#endif //(CAPABILITY&CAP_RESET)
      
   // Assume we have lost connection after reset
   bdm_status.ackn   = WAIT;              // Turn off ACKN feature
   bdm_status.reset  = NO_RESET_ACTIVITY; // BDM resetting the target doesn't count as a reset!
   bdm_rx_ptr        = bdm_rxEmpty;       // Clear the Tx/Rx pointers
   bdm_tx_ptr        = bdm_txEmpty;       //    i.e. no routines found

   if (bdm_status.speed == SPEED_USER_SUPPLIED) {          // User specified speed?
      (void) bdmHC12_confirmSpeed(bdm_status.sync_length); // Confirm we can still operate at that speed
      rc = bdm_enableBDM();                                //  & enable BDM mode
      }
   else if (bdm_option.autoReconnect)           // Re-connect if Auto re-connect enabled
      (void)bdm_connect();                      //    Done even if no SYNC feature - may be slow!
   else     
     bdm_status.speed  = NO_INFO;               // Indicate we no longer have a connection
      
   return rc;
}

//! Interrupt function servicing the IC interrupt from RESET_IN assertion
//!
interrupt //9
void bdm_resetSense(void) {
   CLEAR_RESET_SENSE_FLAG();           // Acknowledge RESET IC Event
   bdm_status.reset = RESET_DETECTED;  // Record the fact that reset was asserted
}

//!  Target power has been externally cycled. Holds RESET/BKGD low while Vdd rises
//!
//! @return
//!     \ref BDM_RC_OK  => Success \n
//!    != \ref BDM_RC_OK     => various errors
//      
U8 bdm_powerOnReset(void) {
U8 rc = 0;

#if (CAPABILITY&CAP_VDDSENSE)

   bdm_idle();  // Make sure BDM interface is idle

   // BKGD pin=L, wait...
   DATA_PORT      = BDM_DIR_Rx_WR;
   DATA_PORT_DDR  = BDM_DIR_Rx_MASK|BDM_OUT_MASK;
//   RESET_LOW();
   WAIT_US(RESET_SETTLE);

   // Release reset
//   RESET_3STATE();

   // Wait for Vdd to rise within 50% of 3V and RESET to return high
   // RESET rise may be delayed by target POR
   WAIT_WITH_TIMEOUT_MS( 250 /* ms */, (bdm_targetVddMeasure()>75)&&
                                     (!bdm_option.useResetSignal)||(RESET_IN!=0));

   // Let signals settle & CPU to finish reset (with BKGD held low)
   WAIT_US(BKGD_WAIT);

   if (bdm_targetVddMeasure()<=70) // Vpp didn't turn on!
      rc = BDM_RC_VDD_NOT_PRESENT;

   if (bdm_option.useResetSignal && (RESET_IN==0)) // RESET didn't rise
      rc = BDM_RC_RESET_TIMEOUT;

   bdm_idle();  // Make sure BDM interface is idle (BKGD now high)

   // Let signals settle
   WAIT_US(RESET_SETTLE);

   if (BDM_IN==0) // BKGD didn't rise!
      rc = BDM_RC_BKGD_TIMEOUT;

   bdm_status.reset = RESET_DETECTED;    // Record the fact that reset was asserted

#endif // (CAPABILITY&CAP_VDDSENSE)
   return(rc);
}

//!  Checks Target Vdd  - Updates Target Vdd LED 
//! 
//!  Updates \ref bdm_status
//!
static void bdm_checkTargetVdd(void) {
#if (CAPABILITY&CAP_VDDSENSE)
   if (bdm_targetVddMeasure() > VDD_2v) {
      RED_LED_ON();
      bdm_status.power = 1;
      }
   else {
      RED_LED_OFF();
      bdm_status.power = 0;
      }
#else
   // No target Vdd sensing - assume Vdd is present
   bdm_status.power = 1;
#endif // CAP_VDDSENSE
}

//! Interrupt function servicing the IC interrupt from Vdd changes
//!
interrupt // 27
void bdm_targetVddSense(void) {

#if (CAPABILITY&CAP_VDDSENSE)
   CLEAR_VDD_SENSE_FLAG(); // Clear Vdd Change Event

   if (VDD_SENSE) {  // Vdd rising
      (void)bdm_powerOnReset(); // Needs to be done on non-interrupt thread?
      bdm_checkTargetVdd();
      }
   else { // Vdd falling
      VDD_OFF();   // Turn off Vdd in case it's an overload
      bdm_idle();  // Make sure BDM interface is idle
      
      // BKGD pin=L, RESET pin=L
      DATA_PORT      = BDM_DIR_Rx_WR;
      DATA_PORT_DDR  = BDM_DIR_Rx_MASK|BDM_OUT_MASK;
      RESET_LOW();
      
      bdm_checkTargetVdd();
      }
#endif // CAP_VDDSENSE
}

//! Waits 64 BDM cycles of the target MCU
//!
void bdm_wait64(void) {
   //enableInterrupts();
   asm {
      ldhx  bdm_status.wait64_cnt    /* number of loop iterations to wait */
   loop:
      aix   #-1      ; [2]
      cphx  #0       ; [3]
      bne   loop     ; [3] 8 cycles / iteration
   }
}

//! Waits 150 BDM cycles of the target MCU
//!
void bdm_wait150(void) {
   //enableInterrupts();
   asm {
      ldhx  bdm_status.wait150_cnt    /* number of loop iterations to wait */
   loop:
      aix   #-1      ; [2]
      cphx  #0       ; [3]
      bne   loop     ; [3] 8 cycles / iteration
   }
}

//! \brief Measures the SYNC length and writes the result into bdm_status structure
//!
//! @return
//!   \ref BDM_RC_OK  => Success \n
//!    \ref BDM_RC_BKGD_TIMEOUT     => BKGD pin stuck low or no response
//!      
U8 bdm_syncMeasure(void) {
   U16 time;

   bdm_status.speed = NO_INFO;  // Indicate that we do not have a clue about target speed at the moment...

   bdm_idle();                  // Make sure BDM interface is idle

   // Wait with timeout until BKGD is high
   WAIT_WITH_TIMEOUT_US(BDM_SYNC_REQ, (BDM_IN!=0));

   if (BDM_IN==0)
      return(BDM_RC_BKGD_TIMEOUT); // timeout !

   // BKGD pin=L, RESET pin=3S (RESET 3-state & pulled high)
   DATA_PORT     = 0           |BDM_DIR_Rx_RD;  
   DATA_PORT_DDR = BDM_OUT_MASK|BDM_DIR_Rx_MASK;  
   DATA_PORT     = 0           |BDM_DIR_Rx_WR;  
   
   WAIT_US(BDM_SYNC_REQ);          // Wait SYNC request time

#if (DEBUG&SYNC_DEBUG)
   DEBUG_PIN   = 0;
   DEBUG_PIN   = 1;
#endif

   // Set up Input capture & timeout timers
   SYNC_TPMSC  = SYNC_TPMSC_FALLING_EDGE_MASK;  // TPM1.Chx : Input capture, falling edge on pin
   TPM1C2SC    = TPM1C2SC_MS2A_MASK;            // TPM1.Ch2 : Output compare (timeout) - no pin used

   TPM1C2V   = TPM1CNT + (U16)(ACKN_MICROSECOND(SYNC_TIMEOUT)); // TPM1.Ch2 : Set time out value (OC)
   TPM1C2SC_CH2F = 0;                                          // TPM1.Ch2 : Clear timeout flag (OC)
   SYNC_TPMSC_CHF = 0;                                         // TPM1.Chx : Clear capture flag

   asm {
      LDX   #BDM_OUT_MASK|BDM_DIR_Rx_WR            // Mask to Drive BKGD high
      LDA   #BDM_OUT_MASK|BDM_DIR_Rx_RD            // Mask to 3-state BKGD
      STX   DATA_PORT                              // [3   wpp]  Drive BKGD high (for 500 ns)
      STA   DATA_PORT                              // [3   wpp]  3-state BKGD
      // It took 125 ns cycles from bringing BKGD high to 3-state (3 cycles/24 MHz)
      // Target drives BKGD line after 16 BDM clock cycles
      // Fast enough up to approx 125ns/(16 BDM cycles) = 128 MHz BDM Frequency
   }

   while ((SYNC_TPMSC_CHF==0)&&(TPM1C2SC_CH2F==0)){   // Wait for capture or timeout
   }
   time        = SYNC_TPMVALUE;                      // TPM1.Ch1 : Save time of start of the SYNC interval
   SYNC_TPMSC  = SYNC_TPMSC_RISINGE_EDGE_MASK;       // TPM1.Ch1 : Input capture, rising edge on pin

   // it takes 23 cycles to re-enable capture (worst case) which is good enough up to 128*24/23 = 130 MHz!
   while ((SYNC_TPMSC_CHF==0)&&(TPM1C2SC_CH2F==0)){   // Wait for capture or timeout
   }
   time   = SYNC_TPMVALUE-time;                  // calculate length of the SYNC

#if (DEBUG&SYNC_DEBUG)
   DEBUG_PIN   = 1;
   DEBUG_PIN   = 0;
#endif

   if (TPM1C2SC_CH2F)
      return(BDM_RC_SYNC_TIMEOUT);         // timeout !

#if (BUS_FREQ==3000000UL)
   bdm_status.sync_length=(time<<2)+(time<<4);  /* multiply by 20 to get the time in 60MHz ticks */
#elif (BUS_FREQ==6000000UL)
   bdm_status.sync_length=(time<<1)+(time<<3);  /* multiply by 10 to get the time in 60MHz ticks */
#elif (BUS_FREQ==24000000UL)
   bdm_status.sync_length=(time<<1)+(time>>1);  // multiply by 2.5 to get the time in 60MHz ticks
#elif (ACKN_TIMER_FREQ==30000000UL)
   bdm_status.sync_length=(time<<1);            // multiply by 2 to get the time in 60MHz ticks
#elif (ACKN_TIMER_FREQ==15000000UL)
   bdm_status.sync_length=(time<<2);            // multiply by 4 to get the time in 60MHz ticks
#else
   #error "Please fix SYNC length calculation for new Bus Frequency"
#endif
   bdm_status.speed = SYNC_SUPPORTED;           // SYNC feature is supported by the target

   return(0);
}

//! Enables ACKN and prepares the timer for easy ACKN timeout use
//!
void bdm_acknInit(void) {

   // Set timer count rate
   ACKN_TPMSC  = ACKN_TPMSC_RISINGE_EDGE_MASK;   // TPM1:Chx : Input capture, rising edge on pin
   TPM1C2SC    = TPM1C2SC_MS2A_MASK;             // TPM1:Ch2 : Output compare - no pin used

#if (DEBUG&ACK_DEBUG)
   DEBUG_PIN     = 0;
   DEBUG_PIN     = 1;
   DEBUG_PIN     = 0;
#endif

   bdm_status.ackn = ACKN;               // Switch ACKN on
   if (bdm_status.target_type==CFV1)
      BDMCF_CMD_ACK_ENABLE();
   else
      BDM_CMD_ACK_ENABLE();              // Send the ACK enable command to the target
}

//! Waits for ACKN pulse from the target (with timeout)
//! If a timeout occurs then the ACKN waiting is disabled
//!
void bdm_ackn(void) {
   TPM1C2V   = TPM1CNT + (U16)(ACKN_MICROSECOND(ACKN_TIMEOUT));   // TPM1.Ch2 : Set time out value (OC)
   TPM1C2SC_CH2F  = 0;                                           // TPM1.Ch2 : Clear timeout flag (OC)

#if (DEBUG&ACK_DEBUG)
   DEBUG_PIN   = 1;
#endif

   //enableInterrupts();
   // Wait for pin capture or timeout
   while ((ACKN_TPMSC_CHF==0)&&(TPM1C2SC_CH2F==0)) { 
   }
   
   if (TPM1C2SC_CH2F) {              // timeout
      bdm_status.ackn = WAIT;        // switch the ackn feature off

#if (DEBUG&ACK_DEBUG)
   DEBUG_PIN   = 0; // Flag ACKN failed
#endif
   }
}

//! Initialises BDM module for the given target processor
//!
//!  @param target = Target processor (see \ref target_type_e)
//!
U8 bdm_init(U8 target) {
U8 rc = 0;

   bdm_off();  // BDM interface is OFF/idle
   
   if (target == T_OFF) {
      bdm_option.targetClockTrim = 0;
      return BDM_RC_OK;
      }
   
   // Set default status/settings
   bdm_status = bdm_statusDefault; 

   bdm_status.target_type  = target;
   
   // TPM1 is set up here and runs continuously for RESET Input Capture, SYNC and AKN measurement
   TPM1SC         = 0;                    // Set timer modulo value (free running)
   TPM1MOD        = 0;                    //  - requires CLKSA:CLKSB=00
   TPM1SC         = ACKN_TPM1SC_VALUE;         // Set timer count rate = BUS Rate

   // TPM2     - General interval timer (no I/O pins)
   TPM2SC         = FAST_TPMxSC_VALUE;    // Set timer count rate
   TPM2C0SC       = TPM2C0SC_MS0A_MASK;   // Output compare function, pin disabled

#if (CAPABILITY&CAP_VDDSENSE)
   //====================================================================
   // Set up Vdd monitoring (ACMP interrupts)
   // Clear existing flag, enable falling & rising transitions (Vdd rising & falling!), enable Bandgap (~1.2V)
   CONFIGURE_VDD_SENSE();         // Capture Vdd rising & falling edges
   CLEAR_VDD_SENSE_FLAG();        // Clear Vdd Change Event
   ENABLE_VDD_SENSE_INT();        // Enable Vdd IC interrupts
#endif

   rc = bdm_cycleTargetVddOn(); // Turn on Vdd if appropriate
   if (rc != BDM_RC_OK)
      return rc;
   
   if (target==RS08)
      FLASH12V_ON(); // Turn on charge pump for RS08 programming

   // Now give the BDM interface enough time to soft reset in case it was doing something before
   WAIT_US(SOFT_RESET);  // Wait for the longest SOFT RESET time possible

   switch (target) {
#if (CAPABILITY&CAP_RESET)
      case HC12:
         bdm_option.useResetSignal = 1; // Must use RESET signal on HC12
         break;
#endif
      case RS08:
         savedTrimValue = 0xFF; // Flag no trim value saved
      case HCS08:
      case CFV1:
         break;

      default:
         return BDM_RC_UNKNOWN_TARGET; // Illegal target set
      }

#if (CAPABILITY&CAP_RESET)
   //====================================================================
   // Setup RESET detection Input Capture
   if (bdm_option.useResetSignal) {
      CONFIGURE_RESET_SENSE();         // Capture RESET falling edges
      CLEAR_RESET_SENSE_FLAG();        // Clear RESET IC Event
      ENABLE_RESET_SENSE_INT();        // Enable RESET IC interrupts
      }
   else
      DISABLE_RESET_SENSE_INT();       // Disable RESET IC interrupts
#endif

   if (bdm_option.cycleVddOnReset && !bdm_option.leaveTargetPowered)
      rc = bdm_cycleTargetVdd();
      
   return rc;
}

//!  Halts the processor (places in background mode)
//!
void bdm_halt(void) {

   if (bdm_status.target_type==CFV1)
      BDMCF_CMD_BACKGROUND();
   else
      BDM_CMD_BACKGROUND();
}

//! Commences full-speed execution on the target
//!
void bdm_go(void) {
U32 x;

   if (bdm_status.target_type == CFV1) {
      // Clear Single-step mode
      BDMCF_CMD_READ_DREG(CFV1_CSR, &x);
      x &= ~CFV1_CSR_SSM;
      BDMCF_CMD_WRITE_DREG(CFV1_CSR,x);
      
      BDMCF_CMD_GO();
      }
   else
      BDM_CMD_GO();
}


//!  Executes a single instruction on the target
//!
void bdm_step(void) {
U32 x;

   if (bdm_status.target_type == CFV1) {
      // Set Single-step mode
      BDMCF_CMD_READ_DREG(CFV1_CSR, &x);
      x |= CFV1_CSR_SSM;
      BDMCF_CMD_WRITE_DREG(CFV1_CSR,x);

      BDMCF_CMD_GO();
      }
   else
      BDM_CMD_TRACE1();
}

//!  Sets the BDM interface to a suspended state
//!
//!  - All signals undriven \n
//!  - All voltages off.
//!
void bdm_suspend(void) {
   VPP_OFF();
   flashState = FLASH_IDLE;
   FLASH12V_OFF();
   VDD_OFF();
   bdm_idle();
}

//!  Turns off the BDM interface
//!
//!  Depending upon settings, may leave target power on.
//!
void bdm_off( void ) {
   flashState = FLASH_IDLE;
   VPP_OFF();
   FLASH12V_OFF();
   if (!bdm_option.leaveTargetPowered)
      VDD_OFF();
   bdm_idle();
}

//!  Sets the BDM hardware interface to an idle condition
//!
//!  BKGD   signal undriven, PUP \n
//!  RESET  signal undriven, PUP \n
//!  VPP,Flash voltage etc unchanged. \n
//!
static void bdm_idle(void) {
   // Enable pull-ups on I/O lines (BKGD & RESET are double pull-ups)
#ifdef DATA_PORT_PEBIT
   // A single bit used to control PUPs on entire port
   DATA_PORT_PEBIT = 1;
#else   
   // Individually controlled PUPs
   DATA_PORT_PER |= BDM_OUT_MASK|BDM_IN_MASK|RESET_OUT_MASK|RESET_IN_MASK|BDM_DIR_Rx_MASK;
#endif   

#ifdef DIR_PORT_PEBIT
   // A single bit used to control PUPs on entire port
   DIR_PORT_PEBIT = 1;
#else   
   // Individually controlled PUPs
   DIR_PORT_PER |= BDM_DIR_MASK;
#endif   

   RESET_3STATE();  
   BDM_3STATE_TX();

   DATA_PORT      = BDM_DIR_Rx_RD;      //  
   DATA_PORT_DDR  = BDM_DIR_Rx_MASK;    // BDM_DIR_Rx enabled on DATA_PORT
   DIR_PORT_DDR  &= ~BDM_DIR_MASK;      // Transfer BKGD 3-state control to BDM_DIR_Rx on DATA_PORT

//   - BKGD & RESET pins on BDM cable are 3-state and may be read through BDM_IN & RESET_IN
//   - BDM_OUT is 3-state
//   - BDM_DIR_Rx is enabled and BDM_DIR is 3-state so BKGD direction is controlled by BDM_DIR_Rx
//   - BDM_DIR_Rx is BDM_DIR_Rx_RD (and hence BKGD pin is 3-state)
//   - RESET_OUT is undriven (but has PUP so is high) (needed if O/C buffer or FET is used as driver on RESET pin)
//   - RESET_DIR is inactive (high or low as required) i.e. external 3-state buffer/transceiver is disabled/input.
}

//! Prepares for transmission of BDM data
//!
//! Interrupts need to be disabled for the duration of all BDM commands. \n
//! It is up to the caller to see to this \n
//!
void bdm_txPrepare(void) {
   //disableInterrupts();

   // Hand over BKGD direction control from BDM_DIR_Rx (on BDM_DATA_PORT) to BDM_DIR (on BDM_DIR_PORT)
   // Note: direction of the BKGD driver must be controlled during transition to avoid glitches on BDM
                                                // Keep BKGD 3-state until transfer
   DATA_PORT     = BDM_OUT_MASK|BDM_DIR_Rx_RD;  // Make sure BKGD is currently 3-state & 
                                                //    will be driven high once the driver is enabled
   BDM_3STATE_TX();                             // Make sure BKGD is 3-state when control is transferred to DIR_PORT
                                                
   DIR_PORT_DDR  |= BDM_DIR_MASK;               // Transfer 3-state control to DIR_PORT
   DATA_PORT_DDR = BDM_OUT_MASK;                // BDM_OUT active others 3-state
   
   // BKGD direction controlled by BDM_DIR in DIR_PORT
   // BDM_DIR is set to disable the BKGD driver
   //    Macros used to enable/disable BKGD drive
   //      BDM_ENABLE_ASM_TX/BDM_ENABLE_TX()            // Enable BKGD
   //      BDM_3STATE_ASM_TX/BDM_3STATE_TX()            // 3-state BKGD         
   // BKGD is driven via BDM_OUT in DATA_PORT - hard coded into the assembly code
   // Other bits of DATA_PORT are configured as inputs and have no effect when written
   // DATA_PORT = 0xxxxxxxx => BKGD driven low  (if enabled by BDM_DIR)
   // DATA_PORT = 1xxxxxxxx => BKGD driven high (if enabled by BDM_DIR)
   // RESET state is unchanged
}

//! Finishes transmission of BDM data and sets up for reception.
//!
void bdm_txFinish(void) {

   // Hand over BKGD direction control from BDM_DIR (on BDM_DIR_PORT) to BDM_DIR_Rx (on BDM_DATA_PORT)
   // Note: direction of the BKGD driver must be controlled during transition to avoid glitches on BDM
   BDM_3STATE_TX();                              // Make sure BKGD is currently 3-state
   DATA_PORT     = BDM_OUT_MASK|BDM_DIR_Rx_RD;   // Make sure BKGD will remain 3-state after handover

   DATA_PORT_DDR  = BDM_OUT_MASK|BDM_DIR_Rx_MASK; // Transfer BDM_DIR control from DIR_PORT to DATA_PORT
   DIR_PORT_DDR  &= ~BDM_DIR_MASK;

   // BKGD is now driven via BDM_OUT (DATA_PORT.7) with 3-state control on BDM_DIR_Rx (DATA_PORT.6).
   // Other bits of DATA_PORT are configured as inputs and have no effect when written.
   // The following examples assume active low enables of the buffer
   // DATA_PORT = 00xxxxxxx => BKGD driven low
   // DATA_PORT = 11xxxxxxx => BKGD 3-state
   // DATA_PORT = 10xxxxxxx => BKGD driven high (not used by Rx routines)
   // DATA_PORT = 01xxxxxxx => BKGD 3-state (not used by Rx routines?)
}

#pragma MESSAGE DISABLE C1404 // Disable warnings about no return statement

//==============================================================
// Tx Routines bdm_tx..
//==============================================================
// Transmit 8 bits of data, MSB first
// These routines assume the following:
//    BKGD direction controlled via BDM_DIR in DIR_PORT
//      (use BDM_3STATE_TX()/BDM_3STATE_ASM_TX or BDM_ENABLE_TX()/BDM_ENABLE_ASM_TX macros)
//    BKGD output value may be controlled by DATA_PORT.7
//    The rest of DATA_PORT must be disabled (call bdm_txPrepare())
// Leaves BDM_OUT 3-state via BDM_DIR (call bdm_txFinish() to clean up)
//
#if (BDM_OUT_BIT != 7)
#error "The bdm_txN(void) routines requires BDM_OUT=PTA.7"
#endif

//=========================================================================
#pragma MESSAGE DISABLE C5703 // Disable warnings about unused parameter
//! Dummy BDM Tx routine
//!
void bdm_txEmpty(U8 data) {
   // If BDM command is executed with this routine set command failed...
   commandBuffer[0] = BDM_RC_NO_RX_ROUTINE;
}

//=========================================================================
// 2,4,6
void bdm_tx1(U8 data) {
   asm {
      LDHX  @DATA_PORT     // Point HX at DATA_PORT
      BDM_ENABLE_ASM_TX    // Enable BKGD (high)

      /* bit 7 (MSB) */
      SEC                  // [1  p   ]   Used by ROR
      CLR   ,X             // [4  rfwp]   Drive BKGD low
                           // --- 6
      STA   ,X             // [2    wp]   Drive data to BDM
                           // --- 2
      ROR   ,X             // [4  rfwp]   Drive BKGD high
                           // --- 4
      LSLA                 // [1  p   ]   Next bit

      /* bit 6 */
      SEC                  // [1  p   ]   Used by ROR
      CLR   ,X             // [4  rfwp]   Drive BKGD low
      STA   ,X             // [2    wp]   Drive data to BDM
      ROR   ,X             // [4  rfwp]   Drive BKGD high
      LSLA                 // [1  p   ]   Next bit

      /* bit 5 */
      SEC                  // [1  p   ]   Used by ROR
      CLR   ,X             // [4  rfwp]   Drive BKGD low
      STA   ,X             // [2    wp]   Drive data to BDM
      ROR   ,X             // [4  rfwp]   Drive BKGD high
      LSLA                 // [1  p   ]   Next bit

      /* bit 4 */
      SEC                  // [1  p   ]   Used by ROR
      CLR   ,X             // [4  rfwp]   Drive BKGD low
      STA   ,X             // [2    wp]   Drive data to BDM
      ROR   ,X             // [4  rfwp]   Drive BKGD high
      LSLA                 // [1  p   ]   Next bit

      /* bit 3 */
      SEC                  // [1  p   ]   Used by ROR
      CLR   ,X             // [4  rfwp]   Drive BKGD low
      STA   ,X             // [2    wp]   Drive data to BDM
      ROR   ,X             // [4  rfwp]   Drive BKGD high
      LSLA                 // [1  p   ]   Next bit

      /* bit 2 */
      SEC                  // [1  p   ]   Used by ROR
      CLR   ,X             // [4  rfwp]   Drive BKGD low
      STA   ,X             // [2    wp]   Drive data to BDM
      ROR   ,X             // [4  rfwp]   Drive BKGD high
      LSLA                 // [1  p   ]   Next bit

      /* bit 1 */
      SEC                  // [1  p   ]   Used by ROR
      CLR   ,X             // [4  rfwp]   Drive BKGD low
      STA   ,X             // [2    wp]   Drive data to BDM
      ROR   ,X             // [4  rfwp]   Drive BKGD high
      LSLA                 // [1  p   ]   Next bit

      /* bit 0 */
      SEC                  // [1  p   ]   Used by ROR
      CLR   ,X             // [4  rfwp]   Drive BKGD low
      STA   ,X             // [2    wp]   Drive data to BDM
      ROR   ,X             // [4  rfwp]   Drive BKGD high

      BDM_3STATE_ASM_TX    // [5      ]   3-state BDM
      ACKN_SETUP_ASM       // [5 rfwpp]   Set up for ACKN
      // ACKN is ready within about 10 bus cycles <500ns
      // Required to be ready 32+16 BDC cycles
      // Adequate for 100MHz+ !
      }
}

//=========================================================================
// 3,4,6
void bdm_tx2(U8 data) {
   asm {
      LDHX  @DATA_PORT     // Point HX at DATA_PORT
      BDM_ENABLE_ASM_TX    // Enable BKGD (high)

      /* bit 7 (MSB) */
      SEC                  // [1  p   ]   Used by ROR
      CLR   ,X             // [4  rfwp]   Drive BKGD low
                           // --- 6
      NOP                  // [1      ]
      STA   ,X             // [2    wp]   Drive data to BDM
                           // --- 3
      ROR   ,X             // [4  rfwp]   Drive BKGD high
                           // --- 4
      LSLA                 // [1  p   ]   Next bit

      /* bit 6 */
      SEC                  // [1  p   ]   Used by ROR
      CLR   ,X             // [4  rfwp]   Drive BKGD low
      NOP                  // [1      ]
      STA   ,X             // [2    wp]   Drive data to BDM
      ROR   ,X             // [4  rfwp]   Drive BKGD high
      LSLA                 // [1  p   ]   Next bit

      /* bit 5 */
      SEC                  // [1  p   ]   Used by ROR
      CLR   ,X             // [4  rfwp]   Drive BKGD low
      NOP                  // [1      ]
      STA   ,X             // [2    wp]   Drive data to BDM
      ROR   ,X             // [4  rfwp]   Drive BKGD high
      LSLA                 // [1  p   ]   Next bit

      /* bit 4 */
      SEC                  // [1  p   ]   Used by ROR
      CLR   ,X             // [4  rfwp]   Drive BKGD low
      NOP                  // [1      ]
      STA   ,X             // [2    wp]   Drive data to BDM
      ROR   ,X             // [4  rfwp]   Drive BKGD high
      LSLA                 // [1  p   ]   Next bit

      /* bit 3 */
      SEC                  // [1  p   ]   Used by ROR
      CLR   ,X             // [4  rfwp]   Drive BKGD low
      NOP                  // [1      ]
      STA   ,X             // [2    wp]   Drive data to BDM
      ROR   ,X             // [4  rfwp]   Drive BKGD high
      LSLA                 // [1  p   ]   Next bit

      /* bit 2 */
      SEC                  // [1  p   ]   Used by ROR
      CLR   ,X             // [4  rfwp]   Drive BKGD low
      NOP                  // [1      ]
      STA   ,X             // [2    wp]   Drive data to BDM
      ROR   ,X             // [4  rfwp]   Drive BKGD high
      LSLA                 // [1  p   ]   Next bit

      /* bit 1 */
      SEC                  // [1  p   ]   Used by ROR
      CLR   ,X             // [4  rfwp]   Drive BKGD low
      NOP                  // [1      ]
      STA   ,X             // [2    wp]   Drive data to BDM
      ROR   ,X             // [4  rfwp]   Drive BKGD high
      LSLA                 // [1  p   ]   Next bit

      /* bit 0 */
      SEC                  // [1  p   ]   Used by ROR
      CLR   ,X             // [4  rfwp]   Drive BKGD low
      NOP                  // [1      ]
      STA   ,X             // [2    wp]   Drive data to BDM
      ROR   ,X             // [4  rfwp]   Drive BKGD high

      BDM_3STATE_ASM_TX    // [5      ]   3-state BDM
      ACKN_SETUP_ASM       // [5 rfwpp]   Set up for ACKN
      }
}

//=========================================================================
// 4,4,6
void bdm_tx3(U8 data) {
   asm {
      LDHX  @DATA_PORT     // Point HX at DATA_PORT
      BDM_ENABLE_ASM_TX    // Enable BKGD (high)

      /* bit 7 (MSB) */
      SEC                  // [1  p   ]   Used by ROR
      CLR   ,X             // [4  rfwp]   Drive BKGD low
                           // --- 6
      NOP                  // [1      ]
      NOP                  // [1      ]
      STA   ,X             // [2    wp]   Drive data to BDM
                           // --- 4
      ROR   ,X             // [4  rfwp]   Drive BKGD high
                           // --- 4
      LSLA                 // [1  p   ]   Next bit

      /* bit 6 */
      SEC                  // [1  p   ]   Used by ROR
      CLR   ,X             // [4  rfwp]   Drive BKGD low
      NOP                  // [1      ]
      NOP                  // [1      ]
      STA   ,X             // [2    wp]   Drive data to BDM
      ROR   ,X             // [4  rfwp]   Drive BKGD high
      LSLA                 // [1  p   ]   Next bit

      /* bit 5 */
      SEC                  // [1  p   ]   Used by ROR
      CLR   ,X             // [4  rfwp]   Drive BKGD low
      NOP                  // [1      ]
      NOP                  // [1      ]
      STA   ,X             // [2    wp]   Drive data to BDM
      ROR   ,X             // [4  rfwp]   Drive BKGD high
      LSLA                 // [1  p   ]   Next bit

      /* bit 4 */
      SEC                  // [1  p   ]   Used by ROR
      CLR   ,X             // [4  rfwp]   Drive BKGD low
      NOP                  // [1      ]
      NOP                  // [1      ]
      STA   ,X             // [2    wp]   Drive data to BDM
      ROR   ,X             // [4  rfwp]   Drive BKGD high
      LSLA                 // [1  p   ]   Next bit

      /* bit 3 */
      SEC                  // [1  p   ]   Used by ROR
      CLR   ,X             // [4  rfwp]   Drive BKGD low
      NOP                  // [1      ]
      NOP                  // [1      ]
      STA   ,X             // [2    wp]   Drive data to BDM
      ROR   ,X             // [4  rfwp]   Drive BKGD high
      LSLA                 // [1  p   ]   Next bit

      /* bit 2 */
      SEC                  // [1  p   ]   Used by ROR
      CLR   ,X             // [4  rfwp]   Drive BKGD low
      NOP                  // [1      ]
      NOP                  // [1      ]
      STA   ,X             // [2    wp]   Drive data to BDM
      ROR   ,X             // [4  rfwp]   Drive BKGD high
      LSLA                 // [1  p   ]   Next bit

      /* bit 1 */
      SEC                  // [1  p   ]   Used by ROR
      CLR   ,X             // [4  rfwp]   Drive BKGD low
      NOP                  // [1      ]
      NOP                  // [1      ]
      STA   ,X             // [2    wp]   Drive data to BDM
      ROR   ,X             // [4  rfwp]   Drive BKGD high
      LSLA                 // [1  p   ]   Next bit

      /* bit 0 */
      SEC                  // [1  p   ]   Used by ROR
      CLR   ,X             // [4  rfwp]   Drive BKGD low
      NOP                  // [1      ]
      NOP                  // [1      ]
      STA   ,X             // [2    wp]   Drive data to BDM
      ROR   ,X             // [4  rfwp]   Drive BKGD high

      BDM_3STATE_ASM_TX    // [5      ]   3-state BDM
      ACKN_SETUP_ASM       // [5 rfwpp]   Set up for ACKN
      }
}

//=========================================================================
// 4,5,8
void bdm_tx4(U8 data) {
   asm {
      BDM_ENABLE_ASM_TX             // Enable BKGD (high)
      LDX   #8

   Loop:
      MOV   #0,DATA_PORT            // [4  pwpp]   Drive BKGD low
                                    // --- 8
      NOP                           // [1      ]
      STA   DATA_PORT               // [3   wpp]   Drive data to BDM
                                    // --- 4
      LSLA                          // [1      ]   Set up next data bit
      MOV   #BDM_OUT_MASK,DATA_PORT // [4  pwpp]   Drive BKGD high
                                    // --- 5
      DBNZX Loop                    // [4      ]   Count bit, finished? no - loop

      BDM_3STATE_ASM_TX             // [5      ]   3-state BDM
      ACKN_SETUP_ASM                // [5 rfwpp]   Set up for ACKN
      }
}

//=========================================================================
// 5,5,8
void bdm_tx5(U8 data) {
   asm {
      BDM_ENABLE_ASM_TX             // Enable BKGD (high)
      LDX   #8

   Loop:
      MOV   #0,DATA_PORT            // [4  pwpp]   Drive BKGD low
                                    // --- 8
      NOP                           // [1      ]
      NOP                           // [1      ]
      STA   DATA_PORT               // [3   wpp]   Drive data to BDM
                                    // --- 5
      LSLA                          // [1      ]   Set up next data bit
      MOV   #BDM_OUT_MASK,DATA_PORT // [4  pwpp]   Drive BKGD high
                                    // --- 5
      DBNZX Loop                    // [4      ]   Count bit, finished? no - loop

      BDM_3STATE_ASM_TX             // [5      ]   3-state BDM
      ACKN_SETUP_ASM                // [5 rfwpp]   Set up for ACKN
      }
}

//=========================================================================
// 5,6,8
void bdm_tx6(U8 data) {
   asm {
      BDM_ENABLE_ASM_TX             // Enable BKGD (high)
      LDX   #8

   Loop:
      MOV   #0,DATA_PORT            // [4  pwpp]   Drive BKGD low
                                    // --- 8
      NOP                           // [1      ]
      NOP                           // [1      ]
      STA   DATA_PORT               // [3   wpp]   Drive data to BDM
                                    // --- 5
      NOP                           // [1      ]
      LSLA                          // [1      ]   Set up next data bit
      MOV   #BDM_OUT_MASK,DATA_PORT // [4  pwpp]   Drive BKGD high
                                    // --- 6
      DBNZX Loop                    // [4      ]   Count bit, finished? no - loop

      BDM_3STATE_ASM_TX             // [5      ]   3-state BDM
      ACKN_SETUP_ASM                // [5 rfwpp]   Set up for ACKN
      }
}

//=========================================================================
// 6,7,8
void bdm_tx7(U8 data) {
   asm {
      BDM_ENABLE_ASM_TX             // Enable BKGD (high)
      LDX   #8

   Loop:
      MOV   #0,DATA_PORT            // [4  pwpp]   Drive BKGD low
                                    // --- 8
      BRN   *+0                     // [3      ]
      STA   DATA_PORT               // [3   wpp]   Drive data to BDM
                                    // --- 6
      NOP                           // [1      ]
      NOP                           // [1      ]
      LSLA                          // [1      ]   Set up next data bit
      MOV   #BDM_OUT_MASK,DATA_PORT // [4  pwpp]   Drive BKGD high
                                    // --- 7
      DBNZX Loop                    // [4      ]   Count bit, finished? no - loop

      BDM_3STATE_ASM_TX             // [5      ]   3-state BDM
      ACKN_SETUP_ASM                // [5 rfwpp]   Set up for ACKN
      }
}

//=========================================================================
// 7,8,8
void bdm_tx8(U8 data) {
   asm {
      BDM_ENABLE_ASM_TX             // Enable BKGD (high)
      LDX   #8

   Loop:
      MOV   #0,DATA_PORT            // [4  pwpp]   Drive BKGD low
                                    // --- 8
      NOP                           // [1      ]
      BRN   *+0                     // [3      ]
      STA   DATA_PORT               // [3   wpp]   Drive data to BDM
                                    // --- 7
      NOP                           // [1      ]
      NOP                           // [1      ]
      NOP                           // [1      ]
      LSLA                          // [1      ]   Set up next data bit
      MOV   #BDM_OUT_MASK,DATA_PORT // [4  pwpp]   Drive BKGD high
                                    // --- 8
      DBNZX Loop                    // [4      ]   Count bit, finished? no - loop

      BDM_3STATE_ASM_TX             // [5      ]   3-state BDM
      ACKN_SETUP_ASM                // [5 rfwpp]   Set up for ACKN
      }
}

//=========================================================================
// 8,9,8
void bdm_tx9(U8 data) {
   asm {
      BDM_ENABLE_ASM_TX             // Enable BKGD (high)
      LDX   #8

   Loop:
      MOV   #0,DATA_PORT            // [4  pwpp]   Drive BKGD low
                                    // --- 8
      NOP                           // [1      ]
      NOP                           // [1      ]
      BRN   *+0                     // [3      ]
      STA   DATA_PORT               // [3   wpp]   Drive data to BDM
                                    // --- 8
      NOP                           // [1      ]
      BRN   *+0                     // [3      ]
      LSLA                          // [1      ]   Set up next data bit
      MOV   #BDM_OUT_MASK,DATA_PORT // [4  pwpp]   Drive BKGD high
                                    // --- 9
      DBNZX Loop                    // [4      ]   Count bit, finished? no - loop

      BDM_3STATE_ASM_TX             // [5      ]   3-state BDM
      ACKN_SETUP_ASM                // [5 rfwpp]   Set up for ACKN
      }
}

//=========================================================================
// 9,10,9
void bdm_tx10(U8 data) {
   asm {
      BDM_ENABLE_ASM_TX             // Enable BKGD (high)
      LDX   #8

   Loop:
      MOV   #0,DATA_PORT            // [4  pwpp]   Drive BKGD low
                                    // --- 9
      BRN   *+0                     // [3      ]
      BRN   *+0                     // [3      ]
      STA   DATA_PORT               // [3   wpp]   Drive data to BDM
                                    // --- 9
      BRN   *+0                     // [3      ]
      NOP                           // [1      ]
      NOP                           // [1      ]
      LSLA                          // [1      ]   Set up next data bit
      MOV   #BDM_OUT_MASK,DATA_PORT // [4  pwpp]   Drive BKGD high
                                    // --- 10
      NOP                           // [1      ]
      DBNZX Loop                    // [4      ]   Count bit, finished? no - loop

      BDM_3STATE_ASM_TX             // [5      ]   3-state BDM
      ACKN_SETUP_ASM                // [5 rfwpp]   Set up for ACKN
      }
}

//=========================================================================
// >=10,>=12,>=18
void bdm_txGeneric(U8 data) {
   asm {
      BDM_ENABLE_ASM_TX             // Enable BKGD (high)
      MOV  #8,bitCount              // # of bits to send

   Loop:
      LDX   txTiming3               // [3      ]
      DBNZX *+0                     // [4n     ]
      MOV   #0,DATA_PORT            // [4  pwpp]   Drive BKGD low
                                    // --- 14+4n   (>=18)
      LDX   txTiming1               // [3      ]
      DBNZX *+0                     // [4n     ]
      STA   DATA_PORT               // [3   wpp]   Drive data to BDM
                                    // ---  6+4n   (>=10)
      LDX   txTiming2               // [3      ]
      DBNZX *+0                     // [4n     ]
      LSLA                          // [1      ]   Next bit
      MOV   #BDM_OUT_MASK,DATA_PORT // [4  pwpp]   Drive BKGD high
                                    // ---  8+4n   (>=12)
      DBNZ  bitCount,Loop           // [7      ]
                                    // ===========
      BDM_3STATE_ASM_TX             // [5      ]   3-state BDM
      ACKN_SETUP_ASM                // [5 rfwpp]   Set up for ACKN
      }
}
#pragma MESSAGE DEFAULT C5703 // Restore warnings about unused parameter


//=================================================================================
// DATA_PORT values for Rx Routines - assumes BDM_DIR_Rx controls BKGD enable/direction
#define RxBDM_LOW         (BDM_DIR_Rx_WR|0)  // BDM_OUT=1 and enabled
#define RxBDM_3_STATE     (BDM_DIR_Rx_RD|0)  // BDM_OUT=0 but 3-state


//==============================================================
//  Rx Routines bdm_rx..
//==============================================================
//  Receive 8 bit of data, MSB first
//  These routines assume the following (call bdm_txFinish()):
//     BDM direction to be controlled by BDM_DIR_Rx in DATA_PORT
//     BDM output value may be controlled by DATA_PORT.7 (hard coded in bdm_rxN())
//     BDM input value may be read from DATA_PORT.0 (hard coded in rx_stackDecode())
//     The rest of PTA must be disabled (configured as inputs)
//  Leaves BDM_OUT 3-state via BDM_DIR_Rx
//  Note: These routines have been re-timed to allow for the phase difference b/w read and write instructions.

#pragma MESSAGE DISABLE C20001 // Disable warnings about stackpointer

//!  Decodes values recorded by RX functions
//!
//!  Expects LSB data in X and remaining 7 bytes on stack
//!  It is expected that caller will JUMP into this routine
//!
static void rx_stackDecode(void) {
   asm {
      MOV   #8,bitCount        // # of bits
decode:
#if (BDM_IN_BIT >= 4)         // Rotate left
#if (BDM_IN_BIT <= 4)
      ROLX                    // Get the interesting bit into C
#endif
#if (BDM_IN_BIT <= 5)
      ROLX                    
#endif
#if (BDM_IN_BIT <= 6)
      ROLX                    
#endif
      ROLX                    
#else // Rotate right
#if (BDM_IN_BIT >= 3)
      RORX                    // Get the interesting bit into C
#endif
#if (BDM_IN_BIT >= 2)
      RORX                    
#endif
#if (BDM_IN_BIT >= 1)
      RORX                    
#endif
      RORX                    
#endif
      RORA                     // and rotate it into A from the top
      PULX                     // get the next value from stack
      DBNZ  bitCount,decode
      PSHX                     // that was one pop too many, so push the value back
   }				
}
#pragma MESSAGE DEFAULT C20001 // Restore warnings about stackpointer

//! Dummy BDM Rx routine
//!
//! When no function appropriate for the target speed can be found the following
//! routine is selected.  This is just to make things safe and to make sure
//! there is a place to jump to in such a case
//!
U8 bdm_rxEmpty( void ) {
   // if BDM command is executed with this routine set command failed...
   commandBuffer[0] = BDM_RC_NO_TX_ROUTINE;
   return(0);
}

#pragma MESSAGE DISABLE C1404  // Disable warnings about no return statement

//=======================================================================
// 3,5,8
U8 bdm_rx1(void) {
#pragma NO_RETURN
   asm {
      LDHX  @DATA_PORT           // Point HX at DATA_PORT

      LDA   #RxBDM_3_STATE       // Writing A to DATA_PORT 3-states BDM
      STA   ,X                   // 3-state BKGD pin

      /* bit 7 (MSB) */
      MOV   #RxBDM_LOW,DATA_PORT // [3  pwp]   Drive BKGD low
                                 // ------- 8
                                 // [+1   p]
      STA   ,X                   // [2   wp]   3-state BKGD pin
                                 // ------- 3
      BRN   *+2                  // [3     ]
      LDA   ,X                   // [2   rf]   Sample BDM_IN
                                 // ------- 5
                                 // [1+   p]
      PSHA                       // [2     ]   Save sample on stack
      LDA   #RxBDM_3_STATE       // [2     ]   Restore A value

      /* bit 6 */
      MOV   #RxBDM_LOW,DATA_PORT
      STA   ,X
      BRN   *+2
      LDA   ,X
      PSHA
      LDA   #RxBDM_3_STATE

      /* bit 5 */
      MOV   #RxBDM_LOW,DATA_PORT
      STA   ,X
      BRN   *+2
      LDA   ,X
      PSHA
      LDA   #RxBDM_3_STATE

      /* bit 4 */
      MOV   #RxBDM_LOW,DATA_PORT
      STA   ,X
      BRN   *+2
      LDA   ,X
      PSHA
      LDA   #RxBDM_3_STATE

      /* bit 3 */
      MOV   #RxBDM_LOW,DATA_PORT
      STA   ,X
      BRN   *+2
      LDA   ,X
      PSHA
      LDA   #RxBDM_3_STATE

      /* bit 2 */
      MOV   #RxBDM_LOW,DATA_PORT
      STA   ,X
      BRN   *+2
      LDA   ,X
      PSHA
      LDA   #RxBDM_3_STATE

      /* bit 1 */
      MOV   #RxBDM_LOW,DATA_PORT
      STA   ,X
      BRN   *+2
      LDA   ,X
      PSHA
      LDA   #RxBDM_3_STATE

      /* bit 0 */
      MOV   #RxBDM_LOW,DATA_PORT
      STA   ,X
      BRN   *+2
      LDX   ,X

      // now get the bit values (last value is in X, previous 7 on stack)
      JMP   rx_stackDecode
   }
}

//=======================================================================
// 3,6,8
U8 bdm_rx2(void) {
#pragma NO_RETURN
   asm {
      LDHX  @DATA_PORT           // Point HX at DATA_PORT

      LDA   #RxBDM_3_STATE       // Writing A to DATA_PORT 3-states BKGD
      STA   ,X                   // 3-state BKGD pin

      /* bit 7 (MSB) */
      MOV   #RxBDM_LOW,DATA_PORT // [3  pwp]   Drive BKGD low
                                 // ------- 8
                                 // [1+   p]
      STA   ,X                   // [2   wp]   3-state BKGD pin
                                 // ------- 3
      BRN   *+2                  // [3     ]
      NOP                        // [1     ]
      LDA   ,X                   // [2   rf]   Sample BDM_IN
                                 // ------- 6
                                 // [1+   p]
      PSHA                       // [2     ]   Save sample on stack
      LDA   #RxBDM_3_STATE       // [2     ]   Restore A value

      /* bit 6 */
      MOV   #RxBDM_LOW,DATA_PORT
      STA   ,X
      BRN   *+2
      NOP
      LDA   ,X
      PSHA
      LDA   #RxBDM_3_STATE

      /* bit 5 */
      MOV   #RxBDM_LOW,DATA_PORT
      STA   ,X
      BRN   *+2
      NOP
      LDA   ,X
      PSHA
      LDA   #RxBDM_3_STATE

      /* bit 4 */
      MOV   #RxBDM_LOW,DATA_PORT
      STA   ,X
      BRN   *+2
      NOP
      LDA   ,X
      PSHA
      LDA   #RxBDM_3_STATE

      /* bit 3 */
      MOV   #RxBDM_LOW,DATA_PORT
      STA   ,X
      BRN   *+2
      NOP
      LDA   ,X
      PSHA
      LDA   #RxBDM_3_STATE

      /* bit 2 */
      MOV   #RxBDM_LOW,DATA_PORT
      STA   ,X
      BRN   *+2
      NOP
      LDA   ,X
      PSHA
      LDA   #RxBDM_3_STATE

      /* bit 1 */
      MOV   #RxBDM_LOW,DATA_PORT
      STA   ,X
      BRN   *+2
      NOP
      LDA   ,X
      PSHA
      LDA   #RxBDM_3_STATE

      /* bit 0 */
      MOV   #RxBDM_LOW,DATA_PORT
      STA   ,X
      BRN   *+2
      NOP
      LDX   ,X

      // now get the bit values (last value is in X, previous 7 on stack)
      JMP   rx_stackDecode
   }
}

//=======================================================================
// 4,6,10
U8 bdm_rx3(void) {
   asm {
      LDA   #0x01                    // Value used as sentinel

   Loop:
      MOV   #RxBDM_LOW,DATA_PORT     // [4  pwpp]   Drive BKGD low
                                     // --- 10
      MOV   #RxBDM_3_STATE,DATA_PORT // [4  pwpp]   3-state BKGD pin
                                     // ---  4
      BRN   *+2                      // [3     ]
      BRSET BDM_IN_BIT,DATA_PORT,*+2 // [3   rpp]   Sample BDM_IN
                                     // ---  6
                                     // [2+   pp]
      ROLA                           // [1      ]   Save sample
      BCC   Loop                     // [3      ]
   }
}

//=======================================================================
// 5,6,10
U8 bdm_rx4(void) {
   asm {
      LDA   #0x01                    // Value used as sentinel

   Loop:
      MOV   #RxBDM_LOW,DATA_PORT     // [4  pwpp]   Drive BKGD low
                                     // --- 10
      NOP                            // [1     ]
      MOV   #RxBDM_3_STATE,DATA_PORT // [4  pwpp]   3-state BKGD pin
                                     // ---  5
      BRN   *+2                      // [3     ]
      BRSET BDM_IN_BIT,DATA_PORT,*+2 // [3   rpp]   Sample BDM_IN
                                     // ---  6
                                     // [2+   pp]
      ROLA                           // [1      ]   Save sample
      BCC   Loop                     // [3      ]
   }
}

//=======================================================================
// 5,8,10
U8 bdm_rx5(void) {
   asm {
      LDA   #0x01                    // Value used as sentinel

   Loop:
      MOV   #RxBDM_LOW,DATA_PORT     // [4  pwpp]   Drive BKGD low
                                     // --- 10
      NOP                            // [1     ]
      MOV   #RxBDM_3_STATE,DATA_PORT // [4  pwpp]   3-state BKGD pin
                                     // ---  5
      BRN   *+2                      // [3     ]
      NOP                            // [1     ]
      NOP                            // [1     ]
      BRSET BDM_IN_BIT,DATA_PORT,*+2 // [3   rpp]   Sample BDM_IN
                                     // ---  8
                                     // [2+   pp]
      ROLA                           // [1      ]   Save sample
      BCC   Loop                     // [3      ]
   }
}

//=======================================================================
// 6,9,10
U8 bdm_rx6(void) {
   asm {
      LDA   #0x01                    // Value used as sentinel

   Loop:
      MOV   #RxBDM_LOW,DATA_PORT     // [4  pwpp]   Drive BKGD low
                                     // --- 10
      NOP                            // [1     ]
      NOP                            // [1     ]
      MOV   #RxBDM_3_STATE,DATA_PORT // [4  pwpp]   3-state BKGD pin
                                     // ---  6
      BRN   *+2                      // [3     ]
      BRN   *+2                      // [3     ]
      BRSET BDM_IN_BIT,DATA_PORT,*+2 // [3   rpp]   Sample BDM_IN
                                     // ---  9
                                     // [2+   pp]
      ROLA                           // [1      ]   Save sample
      BCC   Loop                     // [3      ]
   }
}

//=======================================================================
// >8,>10,>20
U8 bdm_rxGeneric(void) {
   asm {
      LDA   #0x01                    // Value used as sentinel

   Loop:
      LDX   rxTiming1                // [3      ]
      MOV   #RxBDM_LOW,DATA_PORT     // [4  pwpp]   Drive BKGD low
                                     // ---  16+4n
      DBNZX *+0                      // [4n     ]
      MOV   #RxBDM_3_STATE,DATA_PORT // [4  pwpp]   3-state BKGD pin
                                     // ---  4+4n
      LDX   rxTiming2                // [3      ]
      DBNZX *+0                      // [4n     ]
      BRSET BDM_IN_BIT,DATA_PORT,*+2 // [3   rpp]   Sample BDM_IN
                                     // ---  6+4n
                                     // [2+   pp]
      LDX   rxTiming3                // [3      ]
      DBNZX *+0                      // [4n     ]
      ROLA                           // [1      ]   Save sample
      BCC   Loop                     // [3      ]
   }
}

#pragma MESSAGE DEFAULT C1404  // Restore warnings about no return statement

//===============================================================================
//  When the SYNC length expressed in 60MHz ticks is ABOVE OR EQUAL to the value
//  in the table, the corresponding pointer is selected
//  If SYNC is shorter than the second entry, the target runs too fast
//  If SYNC is longer or equal to the last entry, the target runs too slow
//
typedef struct {
   U16   syncThreshold;       // Threshold to use this function
   void  (*txFunc)(U8 data);  // Ptr to selected function
   U8    time1,time2,time3;   // Timing Parameters for function use
} TxConfiguration;

const TxConfiguration txConfiguration[] =
{
{ 0,     bdm_txEmpty,   0, 0, 0 },        //>56 MHz - Max Fequency
{ 137,   bdm_tx1,       1, 0, 0 },        //44 - 56 MHz,       (2,4,6)
{ 167,   bdm_tx2,       2, 0, 0 },        //37.71 - 48 MHz,    (3,4,6)
{ 193,   bdm_tx3,       3, 0, 0 },        //33 - 42 MHz,       (4,4,6)
{ 219,   bdm_tx4,       4, 0, 0 },        //29.33 - 37.33 MHz, (4,5,6)
{ 245,   bdm_tx5,       5, 0, 0 },        //26.4 - 33.6 MHz,   (5,5,6)
{ 270,   bdm_tx6,       6, 0, 0 },        //24 - 30.55 MHz,    (5,6,6)
{ 308,   bdm_tx7,       7, 0, 0 },        //20.31 - 25.85 MHz, (6,7,7)
{ 360,   bdm_tx8,       8, 0, 0 },        //17.6 - 22.4 MHz,   (7,8,7)
{ 412,   bdm_tx9,       9, 0, 0 },        //15.53 - 19.76 MHz, (8,9,8)
{ 463,   bdm_tx10,      10, 0, 0 },       //13.89 - 17.68 MHz, (9,10,9)
{ 527,   bdm_txGeneric, 1, 1, 1 },        //12 - 15.27 MHz,    (10,12,18)
{ 617,   bdm_txGeneric, 2, 1, 1 },        //10.15 - 12.92 MHz, (14,12,18)
{ 720,   bdm_txGeneric, 2, 2, 1 },        //8.8 - 11.2 MHz,    (14,16,18)
{ 824,   bdm_txGeneric, 3, 2, 1 },        //7.76 - 9.88 MHz,   (18,16,18)
{ 927,   bdm_txGeneric, 3, 3, 1 },        //6.95 - 8.84 MHz,   (18,20,18)
{ 1030,  bdm_txGeneric, 4, 3, 2 },        //6.29 - 8 MHz,      (22,20,22)
{ 1181,  bdm_txGeneric, 4, 5, 3 },        //5.28 - 6.72 MHz,   (22,28,26)
{ 1389,  bdm_txGeneric, 5, 6, 4 },        //4.55 - 5.79 MHz,   (26,32,30)
{ 1596,  bdm_txGeneric, 6, 7, 4 },        //4 - 5.09 MHz,      (30,36,30)
{ 1850,  bdm_txGeneric, 7, 9, 6 },        //3.38 - 4.31 MHz,   (34,44,38)
{ 2163,  bdm_txGeneric, 9, 10, 7 },       //2.93 - 3.73 MHz,   (42,48,42)
{ 2520,  bdm_txGeneric, 11, 12, 9 },      //2.49 - 3.17 MHz,   (50,56,50)
{ 2979,  bdm_txGeneric, 12, 16, 11 },     //2.1 - 2.67 MHz,    (54,72,58)
{ 3495,  bdm_txGeneric, 14, 19, 14 },     //1.81 - 2.3 MHz,    (62,84,70)
{ 4057,  bdm_txGeneric, 17, 22, 16 },     //1.55 - 1.98 MHz,   (74,96,78)
{ 4731,  bdm_txGeneric, 22, 24, 19 },     //1.33 - 1.7 MHz,    (94,104,90)
{ 5550,  bdm_txGeneric, 26, 29, 24 },     //1.13 - 1.44 MHz,   (110,124,110)
{ 6514,  bdm_txGeneric, 31, 34, 29 },     //0.96 - 1.23 MHz,   (130,144,130)
{ 7686,  bdm_txGeneric, 37, 40, 34 },     //0.82 - 1.04 MHz,   (154,168,150)
{ 8990,  bdm_txGeneric, 44, 47, 40 },     //0.7 - 0.89 MHz,    (182,196,174)
{ 10529, bdm_txGeneric, 52, 55, 47 },     //0.6 - 0.76 MHz,    (214,228,202)
{ 12298, bdm_txGeneric, 61, 65, 56 },     //0.51 - 0.65 MHz,   (250,268,238)
{ 14501, bdm_txGeneric, 71, 78, 67 },     //0.43 - 0.55 MHz,   (290,320,282)
{ 17084, bdm_txGeneric, 84, 93, 79 },     //0.37 - 0.47 MHz,   (342,380,330)
{ 19718, bdm_txGeneric, 98, 105, 92 },    //0.32 - 0.41 MHz,   (398,428,382)
{ 22627, bdm_txGeneric, 113, 120, 105 },  //0.28 - 0.36 MHz,   (458,488,434)
{ 26068, bdm_txGeneric, 128, 140, 120 },  //0.24 - 0.31 MHz,   (518,568,494)
{ 30170, bdm_txGeneric, 146, 166, 140 },  //0.21 - 0.27 MHz,   (590,672,574)
{ 36571, bdm_txEmpty,   0, 0 ,0  },       //<0.21 MHz - Min. Frequency
};

typedef struct {
   U16   syncThreshold;       // Threshold to use this function
   U8    (*txFunc)(void);     // Ptr to selected function
   U8    time1,time2,time3;   // Timing Parameters for function use
} RxConfiguration;

const RxConfiguration rxConfiguration[] =
{
{ 0,     bdm_rxEmpty,   0, 0, 0 },        //>56 MHz - Max Fequency
{ 137,   bdm_rx1,       1, 0, 0 },        //46.96 - 56 MHz,    (3,5,8)
{ 156,   bdm_rx2,       2, 0, 0 },        //38.57 - 51.43 MHz, (3,6,8)
{ 191,   bdm_rx3,       3, 0, 0 },        //32.73 - 42 MHz,    (4,6,8)
{ 232,   bdm_rx4,       4, 0, 0 },        //28.42 - 33.6 MHz,  (5,6,10)
{ 263,   bdm_rx5,       5, 0, 0 },        //22.5 - 30 MHz,     (5,8,10)
{ 325,   bdm_rx6,       6, 0, 0 },        //18.62 - 24.83 MHz, (6,9,10)
{ 401,   bdm_rxGeneric, 1, 1, 1 },        //14.79 - 19.73 MHz, (8,10,20)
{ 508,   bdm_rxGeneric, 1, 2, 1 },        //11.61 - 15.48 MHz, (8,14,20)
{ 631,   bdm_rxGeneric, 1, 3, 1 },        //9.56 - 12.74 MHz,  (8,18,20)
{ 755,   bdm_rxGeneric, 2, 3, 1 },        //8.12 - 10.83 MHz,  (12,18,20)
{ 879,   bdm_rxGeneric, 2, 4, 2 },        //7.06 - 9.41 MHz,   (12,22,24)
{ 1002,  bdm_rxGeneric, 3, 4, 2 },        //6.24 - 8.32 MHz,   (16,22,24)
{ 1126,  bdm_rxGeneric, 3, 5, 3 },        //5.6 - 7.46 MHz,    (16,26,28)
{ 1305,  bdm_rxGeneric, 5, 5, 5 },        //4.64 - 6.18 MHz,   (24,26,36)
{ 1553,  bdm_rxGeneric, 5, 7, 6 },        //3.96 - 5.27 MHz,   (24,34,40)
{ 1857,  bdm_rxGeneric, 6, 9, 9 },        //3.24 - 4.32 MHz,   (28,42,52)
{ 2230,  bdm_rxGeneric, 8, 10, 11 },      //2.75 - 3.66 MHz,   (36,46,60)
{ 2656,  bdm_rxGeneric, 10, 12, 14 },     //2.28 - 3.04 MHz,   (44,54,72)
{ 3210,  bdm_rxGeneric, 12, 15, 18 },     //1.88 - 2.51 MHz,   (52,66,88)
{ 3884,  bdm_rxGeneric, 15, 18, 23 },     //1.56 - 2.08 MHz,   (64,78,108)
{ 4675,  bdm_rxGeneric, 18, 22, 28 },     //1.3 - 1.73 MHz,    (76,94,128)
{ 5594,  bdm_rxGeneric, 22, 26, 34 },     //1.09 - 1.45 MHz,   (92,110,152)
{ 6687,  bdm_rxGeneric, 27, 31, 42 },     //0.91 - 1.21 MHz,   (112,130,184)
{ 8051,  bdm_rxGeneric, 32, 38, 51 },     //0.75 - 1 MHz,      (132,158,220)
{ 9734,  bdm_rxGeneric, 38, 47, 63 },     //0.62 - 0.83 MHz,   (156,194,268)
{ 11742, bdm_rxGeneric, 45, 58, 77 },     //0.52 - 0.69 MHz,   (184,238,324)
{ 13984, bdm_rxGeneric, 53, 70, 93 },     //0.43 - 0.58 MHz,   (216,286,388)
{ 16905, bdm_rxGeneric, 61, 88, 115 },    //0.36 - 0.48 MHz,   (248,358,476)
{ 20239, bdm_rxGeneric, 72, 108, 140 },   //0.3 - 0.4 MHz,     (292,438,576)
{ 24409, bdm_rxGeneric, 84, 130, 165 },   //0.25 - 0.33 MHz,   (340,526,676)
{ 29028, bdm_rxGeneric, 100, 160, 200 },  //0.21 - 0.28 MHz,   (404,646,816)
{ 36571, bdm_rxEmpty,   0, 0 ,0  },       //<0.21 MHz - Min. Frequency
};

//! Selects Rx and Tx routine to be used according to SYNC length in \ref bdm_status structure
//!  Sets up the soft wait delays
//!
//! @return
//!   \ref BDM_RC_OK  => Success \n
//!   \ref BDM_RC_NO_TX_ROUTINE  => No suitable Tx routine found \n
//!   \ref BDM_RC_NO_RX_ROUTINE  => No suitable Rx routine found \n
//!
U8 bdm_RxTxSelect(void) {
   const TxConfiguration  * far txConfigPtr;
   const RxConfiguration  * far rxConfigPtr;
   
   bdm_rx_ptr = bdm_rxEmpty; // clear the Tx/Rx pointers
   bdm_tx_ptr = bdm_txEmpty; // i.e. no routines found

   for (  txConfigPtr  = txConfiguration+sizeof(txConfiguration)/sizeof(txConfiguration[0]);
        --txConfigPtr >= txConfiguration; ) { // Search the table
        
      if (bdm_status.sync_length >= txConfigPtr->syncThreshold) {     // SYNC is >=
         bdm_tx_ptr    = txConfigPtr->txFunc; // Select this routine
         txTiming1     = txConfigPtr->time1;  // Save timing parameters
         txTiming2     = txConfigPtr->time2;
         txTiming3     = txConfigPtr->time3;
         break;                               // Quit search
      }
   }
   if (bdm_tx_ptr==bdm_txEmpty) // Return if no function found
      return(BDM_RC_NO_TX_ROUTINE);

   for (  rxConfigPtr  = rxConfiguration+sizeof(rxConfiguration)/sizeof(rxConfiguration[0]);
        --rxConfigPtr >= rxConfiguration; ) { // Search the table
        
      if (bdm_status.sync_length >= rxConfigPtr->syncThreshold) {     // SYNC is >=
         bdm_rx_ptr    = rxConfigPtr->txFunc; // Select this routine
         rxTiming1     = rxConfigPtr->time1;  // Save timing parameters
         rxTiming2     = rxConfigPtr->time2;
         rxTiming3     = rxConfigPtr->time3;
         break;                               // Quit search
      }
   }
   if (bdm_tx_ptr==bdm_txEmpty) // Return if no function found
      return(BDM_RC_NO_TX_ROUTINE);

   // Calculate number of iterations for manual delay (each iteration is 8 cycles)
   bdm_status.wait64_cnt  = bdm_status.sync_length/(U16)(((8*60*128UL)/(BUS_FREQ/1000000)/64));
   bdm_status.wait150_cnt = bdm_status.sync_length/(U16)(((8*60*128UL)/(BUS_FREQ/1000000)/150));
   // Correct for overhead in calling function etc. (JSR+RTS+JSR) = (5+4+5) ~ 2 iterations
   if (bdm_status.wait64_cnt<=2)
      bdm_status.wait64_cnt = 1; // minimum of 1 iteration
   else
      bdm_status.wait64_cnt -= 2;
   if (bdm_status.wait150_cnt<=2)
      bdm_status.wait150_cnt = 1; // minimum of 1 iteration
   else
      bdm_status.wait150_cnt -= 2;
   return(0);
}

//! Confirm communication at given Sync value.
//! Only works on HC12 (and maybe only 1 of 'em!)
//!
//! @return
//!   \ref BDM_RC_OK  => Success \n
//!   != \ref BDM_RC_OK  => Various errors
//!
#pragma MESSAGE DISABLE C4001 // Disable warnings about Condition always true
U8 bdmHC12_confirmSpeed(U16 syncValue) {
U8 originalValue;
U8 probe;
U8 rc;

   bdm_status.sync_length = syncValue;

   rc = bdm_RxTxSelect(); // Drivers available for this frequency?
   if (rc != BDM_RC_OK) 
      goto tidyUp;

   rc = BDM_RC_BDM_EN_FAILED; // Assume probing failed

#if 1
   // Get current BDMCCR
   BDM12_CMD_BDREADB(HC12_BDMCCR,&originalValue);
   
   // Set location to 0xAA & read back
   BDM12_CMD_BDWRITEB(HC12_BDMCCR,0xAA);
   BDM12_CMD_BDREADB(HC12_BDMCCR,&probe);
   
   // Read back correctly?
   if (probe != 0xAA)
      goto tidyUp;
      
   // set location to 0x55 & read back
   BDM12_CMD_BDWRITEB(HC12_BDMCCR,0x55);
   BDM12_CMD_BDREADB(HC12_BDMCCR,&probe);
   
   // Read back correctly?
   if (probe != 0x55)
      goto tidyUp;

   // Restore BDMCCR
   BDM12_CMD_BDWRITEB(HC12_BDMCCR,originalValue);
   return BDM_RC_OK; // Success!

#else
   // Get current BDMSTS
   BDM12_CMD_BDREADB(HC12_BDMSTS,&originalValue);
   
   // Try to set BDMSTS.ENBDM
   BDM12_CMD_BDWRITEB(HC12_BDMSTS,originalValue|HC12_BDMSTS_ENBDM);
   BDM12_CMD_BDREADB(HC12_BDMSTS,&probe);
   
   if ((probe & HC12_BDMSTS_ENBDM) == 0) // Not set now? - Try next speed
      goto tidyUp;
      
   // Try to clear BDMSTS.ENBDM
   BDM12_CMD_BDWRITEB(HC12_BDMSTS,originalValue&~HC12_BDMSTS_ENBDM);
   BDM12_CMD_BDREADB(HC12_BDMSTS,&probe);
   
   if ((probe & HC12_BDMSTS_ENBDM) != 0) // Not clear now? - Try next speed
      goto tidyUp;
      
   return BDM_RC_OK; // Success!

#endif   

tidyUp:
   bdm_status.sync_length  = 1;
   bdm_status.speed        = NO_INFO; // Connection cannot be established at this speed
   bdm_status.ackn         = WAIT;    // Clear indication of ACKN feature
   return rc;
}
#pragma MESSAGE DEFAULT C4001 // Disable warnings about Condition always true

//! Attempt to determine target speed by trial and error
//!
//! Basic process used to check for communication is:
//!   -  Attempt to modify the BDM Status register (BDMSTS) or BDM CCR Save Register (BDMCCR)
//!
//! The above is attempted for a range of 'nice' frequencies and then every Tx driver frequency. \n
//! To improve performance the last two successful frequencies are remembered.  This covers the \n 
//! common case of alternating between two frequencies (reset & clock configured)  with a minimum \n
//! number of probes.
//!
static U8 bdmHC12_alt_speed_detect(void) {
static const U16 typicalSpeeds[] = { // Table of 'nice' BDM speeds to try 
   SYNC_MULTIPLE( 8000000UL),  //  8 MHz
   SYNC_MULTIPLE(16000000UL),  // 16 MHz
   SYNC_MULTIPLE( 4000000UL),  //  4 MHz
   SYNC_MULTIPLE(32000000UL),  // 32 MHz
   SYNC_MULTIPLE(24000000UL),  // 24 MHz
   SYNC_MULTIPLE(48000000UL),  // 48 MHz
   SYNC_MULTIPLE(20000000UL),  // 20 MHz
   SYNC_MULTIPLE( 2000000UL),  //  2 MHz
   SYNC_MULTIPLE(10000000UL),  // 10 MHz
   SYNC_MULTIPLE( 1000000UL),  //  1 MHz
   SYNC_MULTIPLE(  500000UL),  // 500kHz
   0
   };
#pragma DATA_SEG __SHORT_SEG Z_PAGE
static U16 lastGuess1 = SYNC_MULTIPLE(8000000UL);  // Used to remember last 2 guesses
static U16 lastGuess2 = SYNC_MULTIPLE(16000000UL); // Common situation to change between 2 speeds (reset,running)
#pragma DATA_SEG DEFAULT
const TxConfiguration  * far txConfigPtr;
int              sub;
U16 currentGuess;
U8 rc;

   // Try last used speed #1
   if (bdmHC12_confirmSpeed(lastGuess1) == BDM_RC_OK) {
      bdm_status.speed = SPEED_GUESSED;  // Speed found by trial and error
      return BDM_RC_OK;
   }

   // Try last used speed #2
   currentGuess = lastGuess2;
   rc = bdmHC12_confirmSpeed(lastGuess2);
   
   
   // Try some likely numbers!
   for (sub=0; typicalSpeeds[sub]>0; sub++) {
      if (rc == BDM_RC_OK)
         break;
      currentGuess = typicalSpeeds[sub];
      rc           = bdmHC12_confirmSpeed(currentGuess);
      }
   
   // Try each Tx driver BDM frequency
   for (  txConfigPtr  = txConfiguration+(sizeof(txConfiguration)/sizeof(txConfiguration[0])-1);
        --txConfigPtr >= txConfiguration; ) { // Search the table
      if (rc == BDM_RC_OK)
         break;
      currentGuess = (txConfigPtr->syncThreshold+(txConfigPtr+1)->syncThreshold)/2;
      rc           = bdmHC12_confirmSpeed(currentGuess);
      }
      

   if (rc == BDM_RC_OK) {
      // Update speed cache (LRU)
      lastGuess2       = lastGuess1;
      lastGuess1       = currentGuess;
      bdm_status.speed = SPEED_GUESSED;  // Speed found by trial and error
      return rc;
   }
   
   bdm_status.speed = NO_INFO;  // Speed not found
   return rc;
}

//! Used to check the phase relationship between port reads and writes.
//!
//!  The higher speed BDM communication routines have very critical timing.  This routine
//!  allowed the timing delays between instructions accessing the ports and port pin changes to be measured.
//!  More precisely, the difference in delays between port reads and writes was measured.
//!
//! @warning This function will hange the BDM
//!
void bdm_checkTiming(void) {

   DIR_PORT_DDR  = 0;     // BKGD direction controlled by DATA_PORT
   DATA_PORT_DDR = BDM_OUT_MASK|BDM_DIR_Rx_MASK;     // BKGD pin is driven, Reset pin is driven

   asm {
      LDHX  @DATA_PORT                               // Point HX at DATA_PORT
      BSET 4,PTFDD                                    // LED pin as output

   Loop1:
      BSET 4,PTFD                                     // [5      ] LED OFF
      BRN  Loop                                       // [3      ]
   Loop:
      LDA  #BDM_OUT_MASK                              // [2      ]
      CLR  0,X                                        // [4  rfwp] Drive BKGD low
                                                      // ---------
      STA  0,X                                        // [2    wp] Drive BKGD high
                                                      // --------- 2 cycles low
      MOV   #BDM_OUT_MASK|BDM_DIR_Rx_MASK,DATA_PORT   // [4  pwpp]   3-state BKGD pin
      NOP                                             // [1      ]
      NOP                                             // [1      ]
      NOP                                             // [1      ]
      NOP                                             // [1      ]
      LDA  0,X                                        // [2    rf] Sample BDM_IN
                                                      // ---------
      BIT  #BDM_IN_MASK                               // [2      ] Input high?
      BEQ  Loop1                                      // [3      ] No - leave LED off

      BCLR 4,PTFD                                     // [5      ] LED on
      BRA  Loop                                       // [3      ]
   }
}

//! Used to check the wait timing
//!
//! @warning This function will hange the BDM.
//!
void bdm_checkWaitTiming(void){

   bdm_idle();  // Make sure BDM interface is idle
   
   for(;;) {
      bdm_wait150();
      DATA_PORT = RxBDM_LOW;
      bdm_wait64();
      DATA_PORT = RxBDM_3_STATE;
   }
}

//!  Measures Target Vdd
//!
//!  @return
//!  8-bit value representing the Target Vdd, N ~ (N/255) * 5V \n
//!  On JM60 hardware the internal ADC is used. 
//!  JB16/UF32 doesn't have an ADC so an external comparator is used.  In this case this routine only
//!  returns an indication if Target Vdd is present (255 => Vdd present, 0=> Vdd not present).
//!
U16 bdm_targetVddMeasure(void) {
   ADCCFG = ADCCFG_ADLPC_MASK|ADCCFG_ADIV1_MASK|ADCCFG_ADIV0_MASK|
            ADCCFG_ADLSMP_MASK|ADCCFG_ADIV1_MASK|ADCCFG_ADIV0_MASK|ADCCFG_ADICLK0_MASK;
   APCTL2 = APCTL2_ADPC9_MASK; // Using channel 9
   ADCSC2 = 0;     // Single software triggered conversion
   ADCSC1 = 9;     // Trigger single conversion on Ch9
   while (ADCSC1_COCO == 0){
   }
   return ADCR;
}

//!  Determines the trim value for the target internal clock.
//!  The target clock is left trimmed to syncValue.
//!
//!  @entry 
//!     Assumes target clock has been suitably configured.
//!  @param      icstrm          Address of clock trim register.
//!  @param      syncValue    SYNC value to trim to (determines frequency, see \ref SYNC_MULTIPLE)
//!
//! @return
//!   \ref BDM_RC_OK  => Success, target clock has been trimmed, and target re-connected at new frequency.
//!                 Global \ref savedTrimValue has calculated trim value \n
//!   != \ref BDM_RC_OK  =>  Various errors \n
//!
U8 bdm_trimTargetClock(U16 icstrm, U16 syncValue){
U8    rc;
U8    temp;
U8    bestTrim;
S16   leastError;
S16   currentError;
#define SEARCH_OFFSET (10)

   // Check if target value set
   if (bdm_option.targetClockTrim == 0)
      return BDM_RC_OK;
      
   savedTrimValue = 0;

   rc = bdm_connect();
   if (rc != BDM_RC_OK)
      return rc;
      
   // Initial binary search
   for (temp = 0x80; temp > 0x0; temp>>=1) {
      savedTrimValue |= temp;
      
      // Set trim value
      BDM08_CMD_WRITEB(icstrm,  savedTrimValue);
//      BDM08_CMD_WRITEB(icstrm+1,   0);
      
      // Measure sync to determine target clock speed      
      rc = bdm_syncMeasure();
      if (rc != BDM_RC_OK)
         return(rc);

      rc = bdm_RxTxSelect();
      if (rc != BDM_RC_OK)
         return(rc);
         
      if (bdm_status.sync_length>syncValue) 
         savedTrimValue &= ~temp; // too slow
   }

   // Assume binary search value is "best" so far
   bestTrim   = savedTrimValue;
   leastError = bdm_status.sync_length-syncValue;
   if (leastError<0)
      leastError = -leastError;
      
   // Linear search +/-SEARCH_OFFSET, starting at higher freq (smaller Trim)
   // Range is constrained to [1..254]
   if (savedTrimValue >= (254-SEARCH_OFFSET))
      temp = 254;
   else
      temp = savedTrimValue + SEARCH_OFFSET;
   if (savedTrimValue <= 1+SEARCH_OFFSET)
      savedTrimValue = 1; 
   else
      savedTrimValue -= SEARCH_OFFSET;
      
   while (savedTrimValue <= temp) {
      // Set trim value
      BDM08_CMD_WRITEB(icstrm,  savedTrimValue);
      
      // measure sync to determine target clock speed      
      rc = bdm_syncMeasure();
      if (rc != BDM_RC_OK)
         return(rc);

      rc = bdm_RxTxSelect();
      if (rc != BDM_RC_OK)
         return(rc);
      
      currentError = bdm_status.sync_length-syncValue;
      if (currentError < 0)
         currentError = -currentError;
      if (currentError < leastError) {
         leastError = currentError;
         bestTrim   = savedTrimValue;
      }
      savedTrimValue++;
   }
   
   savedTrimValue = bestTrim;
   
   // Restore correct Trim value & re-connect
   BDM08_CMD_WRITEB(icstrm,  savedTrimValue);
   (void)bdm_connect();
   
   return BDM_RC_OK;
}

//! Trims the RS08 clock 
//!  @param
//!    syncValue = BDM SYNC value to trim the target clock to (see \ref SYNC_MULTIPLE)
//!
//!  @return
//!   \ref BDM_RC_OK  => Success, target clock has been trimmed, and 
//!          target re-connected. Global \ref savedTrimValue has calculated trim value \n
//!    != \ref BDM_RC_OK => Various errors. Global \ref savedTrimValue has default value \n
//!
U8 bdmRS08_trimTargetClock(U16 syncValue){
U8 rc;

   rc = bdm_connect();
   if (rc == 0) {
      bdm_halt();
      switch (bdm_option.derivative_type) {
         case KA1:
         case KA2:
            // Set up Clock
            BDM08_CMD_WRITEB(RS08KA2_ICSC1,    0);
            BDM08_CMD_WRITEB(RS08KA2_ICSC2,    RS08_ICSV2_BDIV2);
            // Calculate trim
            rc = bdm_trimTargetClock(RS08KA2_ICSTRM, syncValue);
            break;
         case KA4:
         case KA8:
            // Set up Clock
            BDM08_CMD_WRITEB(RS08KA8_ICSC1,    0);
            BDM08_CMD_WRITEB(RS08KA8_ICSC2,    RS08_ICSV2_BDIV2);
            // Calculate trim
            rc = bdm_trimTargetClock(RS08KA8_ICSTRM, syncValue);
            break;
         case LA8:
            // Set up Clock
            BDM08_CMD_WRITEB(RS08LA8_ICSC1,    0);
            BDM08_CMD_WRITEB(RS08LA8_ICSC2,    RS08_ICSV2_BDIV2);
            // Calculate trim
            rc = bdm_trimTargetClock(RS08LA8_ICSTRM, syncValue);
            break;
         default:
            rc = BDM_RC_UNKNOWN_TARGET;
            break;
      }
   }
   if ((rc != BDM_RC_OK) ||                                    // Trim failed
       (savedTrimValue == 0x00) || (savedTrimValue == 0xFF)) { // or resulted in extreme value
      savedTrimValue = 0x80;                                   // replace with 'Safe' trim value
      }

   return rc;
}

#if (CAPABILITY & CAP_FLASH)
//! Turn on Programming Voltage (Vpp = 12V)
//!
//! @return 
//!       \ref BDM_RC_OK               => Success \n
//!       \ref BDM_RC_ILLEGAL_COMMAND  => Illegal command for this target \n
//!    != \ref BDM_RC_OK               => Various errors.
//!
U8 bdmRS08_vppOn( void ) {
U8 rc;

   // Safety check - don't turn on Vpp for wrong target
   if (bdm_status.target_type != RS08) {
      return(BDM_RC_ILLEGAL_COMMAND);
   }
   
   // Make sure we have a target connection
   // This is just trying to check that the target is an RS08
   rc = bdm_connect();
   if (rc != BDM_RC_OK)
      return rc;
   
   VPP_ON();
   
   return BDM_RC_OK;
}
#endif // (CAPABILITY & CAP_FLASH)
