/*! \file
    \brief Header file for BDM routines.
    
 */
#ifndef _BDM_H_
#define _BDM_H_

U8    bdm_init(U8 target);
U8    bdm_RxTxSelect(void);
void  bdm_txPrepare(void);
void  bdm_txFinish(void);
void  bdm_ackn(void);
void  bdm_wait64(void);
void  bdm_wait150(void);
void  bdm_acknInit(void);
U8    bdm_hardwareReset(U8 mode); 
U8    bdm_targetReset( U8 mode);
U8    bdm_syncMeasure(void);
U8    bdm_softwareReset(U8 mode);
U8    bdm_connect(void);
U8    bdm_enableBDM(void);
U8    bdm_readBDMStatus(U8 *status);
U8    bdm_writeBDMControl(U8 bdm_sts);
void  bdm_checkTiming(void);
void  bdm_checkWaitTiming(void);

U8    bdmHC12_confirmSpeed(U16 syncValue);
U8    bdm_makeActiveIfStopped(void);
U8    bdm_cycleTargetVdd(void);
U16   bdm_targetVddMeasure(void);
void  bdm_suspend(void);
void  bdm_off( void );
void  bdm_halt(void);
void  bdm_go(void);
void  bdm_step(void);

U8    bdmRS08_trimTargetClock(U16 syncValue);
U8    bdmRS08_massErase(void);
U8    bdmRS08_programFlash(void);
void  bdmRS08_restoreTrim(void);
U8    bdmRS08_vppOn( void );

//! MACRO to calculate a SYNC value from a given frequency in Hz
#define SYNC_MULTIPLE(x)  (U16)((2*128*(60000000UL/10))/(x/10))

//!  Target microcontroller types
typedef enum {
   HC12  = 0,     //!< HC12 or HCS12 target
   HCS08 = 1,     //!< HCS0-8 target
   RS08  = 2,     //!< RS08 target
   CFV1  = 3,     //!< Coldfire Version 1 target
   T_OFF = 0xFF,  //!< Turn off interface (no target)
} target_type_e;

//!  Reset mode (normal operation or BDM)
typedef enum {	/* type of reset action required */
	SPECIAL_MODE=0,
	NORMAL_MODE=1
} target_mode_e;

//!  Target RS08 microcontroller derivatives
typedef enum {
  KA1=0, //!< RS08KA1
  KA2=1, //!< RS08KA2
  LA8=2, //!< RS08LA8
  KA4=4, //!< RS08KA4
  KA8=5, //!< RS08KA8
} rs08_derivative_type_e;

//! Target supports ACKN or use wait instead
typedef enum {
   ACKN  = 1,   //!< Target supports ACKN feature and it is enabled
   WAIT  = 0,   //!< Use WAIT (delay) instead
} ackn_e;

//! Target reset status values
typedef enum {
   NO_RESET_ACTIVITY    =0,   //!< No reset activiy since last polled
   RESET_DETECTED       =1    //!< Reset since last polled
} reset_e;

//! Target speed selection
typedef enum {
   NO_INFO              =0,   //!< Not connected
   SYNC_SUPPORTED       =1,   //!< Speed determined by SYNC
   SPEED_GUESSED        =2,   //!< Speed determnined by trial & error
   SPEED_USER_SUPPLIED  =3    //!< User has specified the speed to use
} speed_e;

//! BDM Clock source control
typedef enum {
   DEFAULT  = -1,   //!< Use default clock selection (don't modify target's reset default)
   ALT      = 0,    //!< Force ALT clock (CLKSW = 0)
   NORMAL   = 1,    //!< Force Normal clock (CLKSW = 1)
} ClkSwValues;

//! Target status
typedef struct {
   U8  target_type:2;      //!< Target type see \ref target_type_e
   U8  ackn:1;             //!< Target supports ACKN see \ref ackn_e
   U8  reset:1;            //!< Target has been reset, see \ref reset_e
   U8  speed:2;            //!< Target speed determination method, see \ref speed_e
   U8  power:1;            //!< Target Vdd is present
   U16 wait150_cnt;        //!< Time for 150 BDM cycles in bus cycles of the MCU divided by 3
   U16 wait64_cnt;         //!< Time for 64 BDM cycles in bus cycles of the MCU divided by 3
   U16 sync_length;        //!< Length of the target SYNC pulse in 60MHz ticks
} bdm_status_t;

//! Target interface options
typedef struct {
   U8  derivative_type:4;        //!< RS08 Derivative
   U8  targetVdd:2;              //!< Target Vdd (off, 3.3V or 5V)
   U8  cycleVddOnReset:1;        //!< Cycle target Power  when resetting
   U8  cycleVddOnConnect:1;      //!< Cycle target Power if connection problems (when resetting?)
   U8  guessSpeed:1;             //!< Guess speed for target w/o ACKN
   U8  autoReconnect:1;          //!< Automatically re-connect to target (for speed change - experimental)
   U8  leaveTargetPowered:1;     //!< Leave target power on exit
   ClkSwValues useAltBDMClock:2; //!< Use alternative BDM clock source in target (HCS08)
   U8  useResetSignal:1;         //!< Don't use RESET signal on BDM interface
   U16 targetClockTrim;          //!< Clock sync value to trim to (0 indicates no trim)
} bdm_option_t;

//! Flash programming mode
typedef enum {
   FLASH_IDLE    = 0,   //!< No programming in progress, VPP is off
   FLASH_READY   = 2,   //!< Ready to program next block
   FLASH_BUSY    = 4,   //!< Busy programming a block
} flashState_t;

extern void bdm_txEmpty(U8 data);
extern U8   bdm_rxEmpty(void);

#pragma DATA_SEG __SHORT_SEG Z_PAGE
extern U8   (*bdm_rx_ptr)(void); // Pointer to BDM Rx routines
extern void (*bdm_tx_ptr)(U8);   // Pointer to BDM Tx routines

//! Returns true if BDM Rx & Tx routines have been set for the current communication speed
#define BDM_TXRS_SET ((bdm_rx_ptr != bdm_rxEmpty) && (bdm_tx_ptr != bdm_txEmpty))

extern bdm_status_t bdm_status;
extern bdm_option_t bdm_option;

extern volatile U8 flashState;
extern volatile U8 savedTrimValue;

#pragma DATA_SEG DEFAULT

#define MAX_FLASH_PROGRAM (64)  //!< Maximum # of bytes to program into RS08 Flash in a single transaction

//! Buffer for Flash programming
//! 5 byte header + \ref MAX_FLASH_PROGRAM data bytes to program
extern U8 flashProgrammingBuffer[5 + MAX_FLASH_PROGRAM]; 

#define bdm_rx()      (*bdm_rx_ptr)() 
#define bdm_tx(data)  (*bdm_tx_ptr)(data) 

#endif
