/*! \file 
 *  \brief ICP Boot Code
 *
 *  The following code is placed in a region of flash that is not
 *  erased by default.  It detects the need for ICP and also 
 *  redirects the vector table to the user vector table. \n
 *  Loosely Based on Freescale Application Note: AN2399SW
 *
 *  \warning
 *  It is very important that this code does not use any library
 *  routines as they will end up in the USER ROM area which may be replaced. \n
 *  This code must be stand-alone.
 *
    \verbatim
    JM60 ICP Code
    
    Copyright (C) 2008  Peter O'Donoghue


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    \endverbatim

Change History
+====================================================================
|  24 Oct  2008 | Added ICP structure & changes (still compat.) - pgo
|  23 July 2008 | Started coding                                - pgo
+====================================================================
*/

#include <hidef.h>      // Various Macros
#include "derivative.h" // Peripheral declarations
#include "Common.h"
#include "Configure.h"
#include "ICP.h"
#include "ICP_USB.h"
#define USER_FLASH_START (0xC000)                //!< Start of User Area in Flash memory
#define USER_FLASH_END   (FLASH_PROTECT_START-1) //!< End of  User Area in Flash memory

#pragma MESSAGE DISABLE C4301 // Disable warnings about Inline functions
/*
** All this code goes in ICP boot area
*/
#pragma CODE_SEG  BOOT_ROM
#pragma CONST_SEG BOOT_CONST
#pragma DATA_SEG  __SHORT_SEG BOOT_RAM

//! Magic number to indicate ICP reboot
#define MAGIC_REBOOT_KEY (0xAA550FF0UL)

/*! Checked during illegal instruction reset.
**   If set to MAGIC_REBOOT_KEY then ICP boot will occur
*/
static U32 icpKeyBytes;

/*! Flash memory protection boundary
 * 
 *  NVPROT is copied from Flash to Register on reset
 */
static const U8 NVPROT_INIT @0x0000FFBD = ((FLASH_PROTECT_START-1)>>8)&0xFE;

/*! Security options and Vector redirection
 * 
 *  These settings :\n
 *    - Disable security
 *    - Relocate vector table to top of unprotected Flash memory
 *
 *  NVOPT is copied from Flash to Register on reset
 */
static const U8 NVOPT_INIT  @0x0000FFBF = NVOPT_KEYEN_MASK|NVOPT_SEC01_MASK; 

//! Causes reboot to ICP mode
//!
//! Called from user code
//! Doesn't return!
//!
static void rebootToICP(void) {

   icpKeyBytes = MAGIC_REBOOT_KEY;
   CTL = 0x00;   // Disable USB
   reset();            // Reset CPU - doesn't return!
}

/*!  Detects ICP mode
 *
 * ICP mode is indicted by a PTG.1 being low
**   OR
** Software initiated ICP by illegal instruction reset + correct magic number
**
 * This is failsafe ICP option during development.  
 * PTG.1 is not accessible in final version (shrink wrapped).
 */
static char detectICP(void) {
#define TSTPIN (1<<1)  // Test output
U8 doICP;

   // Forced Reboot to ICP 
   // Indicated by Illegal instruction reset + correct Magic Number
   if (SRS_ILOP && (icpKeyBytes == MAGIC_REBOOT_KEY)) {
      icpKeyBytes = 0;  // No more ICP reboots
      return 1;
   }
   icpKeyBytes = 0;  // No more ICP reboots
   
   ICP_PIN_PER = 1;    // Enable pull-up
   ICP_PIN_DDR = 0;    // Make input
   asm {
      nop
      nop
   }
   doICP     = !ICP_PIN;    // Check pin
   
   // Restore reset state
   ICP_PIN_PER = 0;         // Disable pull-up
   
   return doICP;
}

/*! Check if user region of flash is invalid
 *
 * First checks for invalid reset vector then does a simple checksum over the user flash.
 * See PRM file for example of automatic calculation of checksum needed
 */
#pragma MESSAGE DISABLE C1404 // Disable warnings about return expected
static U8 flashInvalid(){
   asm{
      lda      #1                     // Set up for failed test
      
      // Check for valid user reset vector
      ldhx     userVectorTable:0      // Reset vector in user vector table

      cphx     #USER_FLASH_START      // Exit if not pointing within flash
      blo      ex
      cphx     #USER_FLASH_END
      bhi      ex

      // Do checksum over user flash memory
      ldhx     #USER_FLASH_START
      clra
      
      // Sum the flash memory
   loop:
      add      ,x
      aix      #1
      cphx     #USER_FLASH_END
      blo      loop
      
      sub      ,x   // Subtract the checksum
                    // 0 => OK
   ex:
   }
}
#pragma MESSAGE DEFAULT C1404 // Restore warnings about return expected

#pragma NO_RETURN
#pragma MESSAGE DISABLE C1404 // Disable warnings about return expected
//!  Calls user ICP detection routine via fixed location vector
//!
static U8 _userDetectICP(void){
   asm {
      ldhx ICP_data.userDetectICP
      jmp  0,x
   }
}
#pragma MESSAGE DEFAULT C1404 // Restore warnings about return expected

#pragma NO_RETURN
/*! Checks if ICP is required
 *   If so, ICP code in ROM is called
 *   otherwise user reset code is called via user vector table
 *  Most of the action is driven from the USB polling function.  
 *
 * @return Never exits
 */
static void myStartup(void) {

   asm {
   ldhx  #$1FF  // Need to set realistic stack
   txs
   }
   LED_INIT();
   if (detectICP() || flashInvalid() || _userDetectICP()) {
      SOPT1  = SOPT1_STOPE_MASK; // Disable COP & enable STOP instruction
      initICP_USB();             // Initiliase minimum USB stack
      USBEventPollingLoop();     // Poll USB events forever
   }
   // Jmp to user code via user reset vector
   userVectorTable[0]();
}
//! Vector for rebootToICP() routine
static const vector forceICPVector @ICP_FORCE_LOCATION = (vector)rebootToICP;

//! Reset Vector
//! All other vectors are re-mapped
static const vector resetVector @0xFFFE = myStartup;

#pragma CONST_SEG DEFAULT
//! ICP version (2 hex digits, major.minor)
#define ICP_VERSION_SW (1<<4|2) // 1.2
 
const U8 ICP_Version_SW @ ICP_VERSION_SW_LOCATION = ICP_VERSION_SW; 
const U8 ICP_Version_HW @ ICP_VERSION_HW_LOCATION = VERSION_HW;     
